from typing import List, Dict, Union, Any, Tuple

from nexinfosys.command_executors import BasicCommand, CommandExecutionError, subrow_issue_message
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.command_generators import IType
from nexinfosys.command_generators.parser_ast_evaluators import ast_evaluator, dictionary_from_key_value_list
from nexinfosys.command_generators.parser_field_parsers import string_to_ast, expression_with_parameters
from nexinfosys.common.helper import strcmp, first, FloatOrString
from nexinfosys.model_services import State
from nexinfosys.models.musiasem_concepts import Processor, ProcessorsRelationPartOfObservation, \
    FactorsRelationScaleObservation, ProcessorsSet


class ProcessorScalingsCommand(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))

    def _process_row(self, fields: Dict[str, Any], subrow=None) -> None:
        scaling_type = fields["scaling_type"]
        scale: str = fields["scale"]

        # Find processors
        invoking_processor = self._get_processor_from_field("invoking_processor")
        requested_processor = self._get_processor_from_field("requested_processor")

        if invoking_processor == requested_processor:
            raise CommandExecutionError(f"Invoking and Requested processors cannot be the same '{invoking_processor.name}'. "
                                        f"Use the 'relative_to' attribute in 'Interfaces' command instead.")

        invoking_interface_name: str = fields["invoking_interface"]
        requested_interface_name: str = fields["requested_interface"]

        requested_new_processor_name: str = fields["new_processor_name"]

        ##
        # Transform text of "attributes" into a dictionary
        if fields.get("attributes"):
            try:
                fields["attributes"] = dictionary_from_key_value_list(fields["attributes"], self._glb_idx)
            except Exception as e:
                self._add_issue(IType.ERROR, str(e)+subrow_issue_message(subrow))
                return
        else:
            fields["attributes"] = {}

        # Process specific fields

        # Obtain the parent: it must exist. It could be created dynamically but it's important to specify attributes
        if fields.get("parent_processor"):
            try:
                parent_processor = self._get_processor_from_field("parent_processor")
            except CommandExecutionError:
                self._add_issue(IType.ERROR, f"Specified parent processor, '{fields.get('parent_processor')}', does not exist"+subrow_issue_message(subrow))
                return
        else:
            parent_processor = None

        # Get internal and user-defined attributes in one dictionary
        attributes = {c.name: fields[c.name] for c in self._command_fields if c.attribute_of == Processor and fields[c.name] is not None}

        # print(f"Invoking: {invoking_processor.name}:{invoking_interface_name}, Requested: {requested_processor.name}:{requested_interface_name}")

        requested_processor_clone = None
        if strcmp(scaling_type, "CloneAndScale") or strcmp(scaling_type, "Clone"):
            # TODO: check “RequestedProcessor” must be an archetype
            # 1. Clones “RequestedProcessor” as a child of “InvokingProcessor”
            requested_processor_clone = self._clone_processor_as_child(processor=requested_processor,
                                                                       parent_processor=invoking_processor if not parent_processor else parent_processor,
                                                                       name=requested_new_processor_name,
                                                                       other_attributes=attributes)

            if strcmp(scaling_type, "CloneAndScale"):
                # 2. Constrains the value of “RequestedInterface” to the value of “InvokingInterface”, scaled by “Scale”
                try:
                    self._constrains_interface(scale=scale,
                                               invoking_interface_name=invoking_interface_name,
                                               requested_interface_name=requested_interface_name,
                                               parent_processor=invoking_processor,
                                               child_processor=requested_processor_clone)
                except Exception as e:
                    self._add_issue(IType.ERROR, str(e)+subrow_issue_message(subrow))
                    return
        elif strcmp(scaling_type, "Scale"):
            # Processors must be of same type (archetype or instance)
            if not strcmp(invoking_processor.instance_or_archetype, requested_processor.instance_or_archetype):
                raise CommandExecutionError("Requested and invoking processors should be of the same type "
                                            "(both instance or_archetype)")

            # 1. Constrains the value of “RequestedInterface” to the value of “InvokingInterface”, scaled by “Scale”
            try:
                self._constrains_interface(scale=scale,
                                           invoking_interface_name=invoking_interface_name,
                                           requested_interface_name=requested_interface_name,
                                           parent_processor=invoking_processor,
                                           child_processor=requested_processor)
            except Exception as e:
                self._add_issue(IType.ERROR, str(e)+subrow_issue_message(subrow))
                return

        elif strcmp(scaling_type, "CloneScaled"):
            # “RequestedProcessor” must be an archetype
            # if not strcmp(requested_processor.instance_or_archetype, "archetype"):
            #     raise CommandExecutionError(f"Requested processor '{requested_processor.name}' should be of type 'archetype'")

            # “InvokingProcessor” must be an instance
            # if not strcmp(invoking_processor.instance_or_archetype, "instance"):
            #     raise CommandExecutionError(f"Invoking processor '{invoking_processor.name}' should be of type 'instance'")

            # 1. Clones “RequestedProcessor” as a child of “InvokingProcessor”
            # 2. Scales the new processor using “Scale” as the value of “RequestedInterface”
            requested_processor_clone = self._clone_processor_as_child(processor=requested_processor,
                                                                       parent_processor=invoking_processor if not parent_processor else parent_processor,
                                                                       other_attributes=attributes)

            # Value Scale, which can be an expression, should be evaluated (ast) because we need a final float number
            scale_value = self._get_scale_value(scale)

            # In the cloned processor search in all interfaces if there are Observations relative_to RequestedInterface
            # and multiply the observation by the computed scale.
            self._scale_observations_relative_to_interface(processor=requested_processor_clone,
                                                           interface_name=requested_interface_name,
                                                           scale=scale_value)

        if requested_processor_clone:
            # Find or create processor and REGISTER it in "glb_idx"
            # Add to ProcessorsGroup, if specified
            field_val = fields.get("processor_group")
            if field_val:
                p_set = self._p_sets.get(field_val, ProcessorsSet(field_val))
                self._p_sets[field_val] = p_set
                if p_set.append(requested_processor_clone,
                                self._glb_idx):  # Appends codes to the pset if the processor was not member of the pset
                    p_set.append_attributes_codes(fields["attributes"])

    def _clone_processor_as_child(self, processor: Processor, parent_processor: Processor, name: str = None,
                                  other_attributes: Dict = {}) -> Processor:
            # Clone inherits some attributes from parent
            inherited_attributes = dict(
                subsystem_type=parent_processor.subsystem_type,
                processor_system=parent_processor.processor_system,
                instance_or_archetype=parent_processor.instance_or_archetype
            )

            processor_clone, processor_clone_children = processor.clone(state=self._glb_idx, name=name,
                                                                        inherited_attributes={**inherited_attributes,
                                                                                              **other_attributes}
                                                                        )

            # Create PART-OF relation
            relationship = ProcessorsRelationPartOfObservation.create_and_append(parent=parent_processor,
                                                                                 child=processor_clone)
            self._glb_idx.put(relationship.key(), relationship)

            # Add cloned processor hierarchical names to global index
            Processor.register([processor_clone] + list(processor_clone_children), self._glb_idx)

            return processor_clone

    def _constrains_interface(self,
                              scale: str,
                              invoking_interface_name: str,
                              requested_interface_name: str,
                              parent_processor: Processor,
                              child_processor: Processor):

        origin_factor = first(parent_processor.factors, lambda i: strcmp(i.name, invoking_interface_name))

        if not origin_factor:
            raise Exception("Invoking interface name '"+invoking_interface_name+"' not found for processor '"+parent_processor.name+"'")

        destination_factor = first(child_processor.factors, lambda i: strcmp(i.name, requested_interface_name))

        if not destination_factor:
            raise Exception("Requested interface name '"+invoking_interface_name+"' not found for processor '"+parent_processor.name+"'")

        if origin_factor.taxon != destination_factor.taxon:
            # Search for an Interface Type Conversion defined in the ScaleChangeMap command
            interface_types_transform = self._get_interface_types_transform(origin_factor.taxon, parent_processor,
                                                                            destination_factor.taxon, child_processor)
            scale = FloatOrString.multiply(scale, interface_types_transform.scaled_weight)

        relationship = FactorsRelationScaleObservation.create_and_append(origin=origin_factor,
                                                                         destination=destination_factor,
                                                                         observer=None,
                                                                         quantity=scale)

        # relationship = ProcessorsRelationUpscaleObservation.create_and_append(parent=parent_processor,
        #                                                                       child=child_processor,
        #                                                                       observer=None,
        #                                                                       factor_name=interface_name,
        #                                                                       quantity=scale)

        self._glb_idx.put(relationship.key(), relationship)

    def _get_scale_value(self, scale: str):
        try:
            value = float(scale)
        except ValueError:
            ast = string_to_ast(expression_with_parameters, scale)

            evaluation_issues: List[Tuple[int, str]] = []
            s = State()
            value, unresolved_vars = ast_evaluator(exp=ast, state=s, obj=None, issue_lst=evaluation_issues)

            if len(evaluation_issues) > 0:
                evaluation_issues_str = [i[1] for i in evaluation_issues]
                raise CommandExecutionError(f"Problems evaluating scale expression '{scale}': "
                                            f"{', '.join(evaluation_issues_str)}")
            elif len(unresolved_vars) > 0:
                raise CommandExecutionError(f"Unresolved variables evaluating the scale expression '{scale}':"
                                            f" {', '.join(unresolved_vars)}")

            elif not value:
                raise CommandExecutionError(f"The scale expression '{scale}' could not be evaluated.")

        return value

    def _scale_observations_relative_to_interface(self, processor: Processor, interface_name: str,
                                                  scale: Union[int, float]):
        for factor in processor.factors:
            for observation in factor.quantitative_observations:
                relative_to_interface = observation.attributes.get("relative_to", None)
                if relative_to_interface and strcmp(relative_to_interface.name, interface_name):
                    observation.value = float(observation.value) * scale
                    observation.attributes["relative_to"] = None




"""
Export processors to XML
Inputs are both the registry and the output dataframe
The registry serves to prepare the structure of the file

"""
from typing import Dict, Tuple

from nexinfosys import case_sensitive
from nexinfosys.common.helper import strcmp, PartialRetrievalDictionary
from nexinfosys.model_services import State, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import ProcessorsRelationPartOfObservation, Processor, Factor


def export_model_to_xml(registry: PartialRetrievalDictionary) -> Tuple[str, Dict[str, Processor]]:
    """
    Elaborate an XML string containing the nested processors and their attributes.
    Also the interfaces inside processors
    <processors>
      <root_p1 fullname="" level="" system="" subsystem="" functional="true|false">
        <interfaces>
          <i1 type="" sphere="" roegen_type="" orientation="" opposite_processor_type="" />
          ...
        </interfaces>
        <child_p2>
          ...
        </child_p2>
      </root_p1>
      ...
    </processors>

    Example (abstract):

    '/processors//[level="n"]'

    :param registry:
    :return:
    """

    def xml_processor(p: Processor, registry: PartialRetrievalDictionary, p_map: Dict[str, Processor]):
        """
        Return the XML of a processor
        Recursive into children

        :param p:
        :return:
        """

        def xml_interface(iface: Factor):
            """

            :param iface:
            :return:
            """
            s = f'<{iface.name} type="{iface.taxon.name}" sphere="{iface.sphere}" ' \
                f'roegen_type="{iface.roegen_type}" orientation="{iface.orientation}" ' \
                f'opposite_processor_type="{iface.opposite_processor_type}" />'
            if case_sensitive:
                return s
            else:
                return s.lower()

        children = p.children(registry)
        full_name = p.full_hierarchy_names(registry)[0]
        if case_sensitive:
            p_map[full_name] = p
        else:
            p_map[full_name.lower()] = p

        s = f"""
    <{p.name} fullname="{full_name}" level="{p.level}" system="{p.processor_system}" subsystem="{p.subsystem_type}" functional="{"true" if strcmp(p.functional_or_structural, "Functional") else "false"}" >
        <interfaces>
        {chr(10).join([xml_interface(f) for f in p.factors])}
        </interfaces>
        {chr(10).join([xml_processor(c, registry, p_map) for c in children])}    
    </{p.name}>"""
        if case_sensitive:
            return s
        else:
            return s.lower()

    # Part of relationships
    por = registry.get(ProcessorsRelationPartOfObservation.partial_key())

    # Set of all instance processors NOT touched by part-of relationships
    unaffected_procs = set([p for p in registry.get(Processor.partial_key()) if strcmp(p.instance_or_archetype, "Instance")])
    for po in por:
        try:
            unaffected_procs.remove(po.parent_processor)
        except KeyError:
            pass
        try:
            unaffected_procs.remove(po.child_processor)
        except KeyError:
            pass

    # Keep those affecting Instance processors
    por = [po for po in por if strcmp(po.parent_processor.instance_or_archetype, "Instance")]

    # Get root processors (set of processors not appearing as child_processor)
    parents = set([po.parent_processor for po in por])
    children = set([po.child_processor for po in por])
    roots = parents.difference(children).union(unaffected_procs)
    # leaves = children.difference(parents)
    result = '<processors>'  # <?xml version="1.0" encoding="utf-8"?>\n
    p_map = {}
    for p in roots:
        result += xml_processor(p, registry, p_map)
    result += "\n</processors>"

    return result, p_map



import json
import logging

from nexinfosys.command_executors import create_command


def commands_generator_from_native_json(input, state, sublist, stack):
    """
    It allows both a single command ("input" is a dict) or a sequence of commands ("input" is a list)

    :param input: Either a dict (for a single command) or list (for a sequence of commands)
    :param state: State used to check variables
    :param sublist: List of command names to consider. If not defined, ALL commands
    :param stack: Stack of nested files. Just pass it...
    :return: A generator of IExecutableCommand
    """
    def build_and_yield(d):
        # Only DICT is allowed, and the two mandatory attributes, "command" and "content"
        if isinstance(d, dict) and "command" in d and "content" in d:
            if "label" in d:
                n = d["label"]
            else:
                n = None
            logging.debug(f"{d['command']} build and yield")
            yield create_command(d["command"], n, d["content"])

    json_commands = json.loads(input)  # Convert JSON string to dictionary
    if isinstance(json_commands, list):  # A sequence of primitive commands
        for i in json_commands:  # For each member of the list
            yield from build_and_yield(i)
    else:  # A single primitive command
        yield from build_and_yield(json_commands)



from openpyxl.worksheet.worksheet import Worksheet

from nexinfosys import AreaTupleType, IssuesLabelContentTripleType
from nexinfosys.command_generators import parser_field_parsers


def parse_hierarchy_command(sh: Worksheet, area: AreaTupleType, name: str, n_type: str) -> IssuesLabelContentTripleType:
    """
    Analyze a "hierarchy" command expressed in a worksheet of a spreadsheet

    The resulting JSON will be:
    {
    "name": <hierarchy name>,
    "type": ("Category", "FactorType", "Processor"), (this determines if the hierarchy is "is-a" -categories or factor types- or "part-of" -processors-)
    "h": [{"name": ..., "description": ..., "expression": ..., children: []},
         ]
    }

    In a hierarchy only simple names (not hierarchic) are allowed. The full name is determined by its position in the tree
    At execution time, if the elements already exist, their location in the hierarchy is updated (and the description, if present, is added)

    :param sh: Input worksheet
    :param area: Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the
    command is present
    :param n_type: Type of hierarchy node: "C" (Category), "I" (InterfaceType) or "P" (Processor)
    :return: list of issues [(issue_type, message)], command label, command content
    """
    some_error = False
    issues = []

    col_names = {("expression", "formula"): "expression",
                 ("code", "name"): "code",
                 ("description",): "description"
                 }

    # Scan columns to prepare:
    # * "expression_column". The column that can contain an expression (it is Optional)
    # * "levels". List of Levels, formed by pairs "code, description", where "description" is optional
    expression_column = None
    levels = []
    for c in range(area[2], area[3]):  # Scan all columns
        col_name = sh.cell(row=area[0], column=c).value
        if not col_name:
            continue

        for k in col_names:
            col_name = col_name.lower()
            if col_name in k:
                if col_name == "expression":
                    expression_column = c
                elif col_name == "code":
                    levels.append(tuple([c]))
                elif col_name == "description":
                    # Description if there is an active CODE. If the description for the active CODE was
                    # already satified, replace it...
                    if len(levels) > 0:
                        tmp = levels[-1]
                        levels[-1] = (tmp[0], c)  # Code, Description
                break

    # Now, scan rows.
    # Only one Level can be active at a time.
    # Current level starts in zero, and is updated in each row.
    # Level can increase by one with regard to the previous level, or freely decrease
    nodes = {}  # Store nodes to check expressions later. Row number is key of the dictionary, the node is the value
    nodes_stack = []
    current_level = -1
    for r in range(area[0]+1, area[1]):
        found = False
        for level, t in enumerate(levels):
            code_column = t[0]

            value = sh.cell(row=r, column=code_column).value
            if value:
                found = True
                break
        if found:
            # Value syntax. A simple identity name
            try:
                parser_field_parsers.string_to_ast(parser_field_parsers.simple_ident, value)
            except:
                issues.append((3, "The name of the category must be a simple name. Row "+str(r)))
            # Description
            if len(t) > 1:
                description_column = t[1]
                description = sh.cell(row=r, column=description_column).value
            else:
                description = None
            # Expression
            if expression_column:
                expression = sh.cell(row=r, column=expression_column).value
            else:
                expression = None
            # Create the hierarchy node
            n = dict(code=value, description=description, expression=expression, children=[])
            if not n["expression"]:
                del n["expression"]
            if not n["description"]:
                del n["description"]
            # Store the node
            nodes[r] = n

            # Process hierarchical information
            add_node = True
            if level == current_level + 1:
                # New (empty) list
                nodes_stack.append([])
                current_level = level
            elif level <= current_level:
                while current_level > level:
                    lst = nodes_stack.pop()  # Take and remove last element of the stack
                    current_level -= 1
                    if current_level >= 0:
                        # From the current level, children of the last node of the list are defined in "lst"
                        nodes_stack[current_level][-1]["children"] = lst
            else:
                issues.append((3, "Hierarchical level must increase by one, not more. Previous level was "+str(current_level)+", current is "+str(level)+". Row "+str(r)))
                add_node = False
            # Append the new node to the current level
            if add_node:
                nodes_stack[current_level].append(n)

    # Close
    while current_level > 0:
        lst = nodes_stack.pop()  # Take and remove last element of the stack
        current_level -= 1
        if current_level >= 0:
            # From the current level, children of the last node of the list are defined in "lst"
            nodes_stack[current_level][-1]["children"] = lst
    # Check that expressions are correct and that they refer to existing codes
    # TODO Check that expressions are not circular
    codes = set([n["code"].lower() for n in nodes.values()])
    for r, n in nodes.items():
        code = n["code"]
        if "expression" in n:
            expression = n["expression"]
            ast = parser_field_parsers.string_to_ast(parser_field_parsers.hierarchy_expression, expression)
            for p in ast["terms"]:
                if isinstance(p, str):
                    if p.lower() not in codes:
                        issues.append((3, "The code '"+p+"' in the expression '"+expression +
                                          "' (declaration of code '"+code+"') was not defined. Row: "+str(r)))

    content = {"name": name,
               "type": n_type,
               "h": nodes_stack[0]}

    return issues, None, content





from openpyxl.worksheet.worksheet import Worksheet

from nexinfosys import IssuesLabelContentTripleType, AreaTupleType
from nexinfosys.command_generators import parser_field_parsers


def parse_pedigree_matrix_command(sh: Worksheet, area: AreaTupleType, name: str) -> IssuesLabelContentTripleType:
    """
    A pedigree matrix is formed by several columns, with a header naming a phase, and below a list of modes, normally in
    ascending qualitative order.

    Modes can be referred later by the order number specified in the "Code" column (mandatory). The order of the columns
    serves also to sequence the codes of the matrix, from left to right.

    Columns can be accompanied by a description column, to the right

    :param sh: Input worksheet
    :param area: Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the
    command is present
    :param name: Name of the Pedigree Matrix
    :return: list of issues (issue_type, message), command label, command content
    """

    issues = []

    # Analyze columns
    phases = []  # A phase per column
    codes = None  # Column with codes
    max_len = 0  # Column with max length
    for c in range(area[2], area[3]):
        phase_modes = []
        current_phase = None
        for r in range(area[0], area[1]):
            value = sh.cell(row=r, column=c).value
            # First row has to be defined. If not, skip to the next column
            if r == area[0] and not value:
                break
            if value is None:
                continue

            if r == area[0]:
                current_phase = value

            try:
                if current_phase.lower() != "code":
                    parser_field_parsers.string_to_ast(parser_field_parsers.simple_ident, value)
                else:
                    if r != area[0]:
                        # An Integer
                        try:
                            int(value)
                        except:
                            issues.append((3, "The code must be an integer"))
            except:
                if r == area[0]:
                    issues.append((3, "Phase '"+value+"' of the Pedigree Matrix must be a simple identity (alphabet letter followed by either alphabet letters or numbers"))
                else:
                    issues.append((3, "A mode ("+value+") in phase '"+current_phase+"' of the Pedigree Matrix must be a simple identity (alphabet letter followed by either alphabet letters or numbers"))

            # Append mode to the current phase
            phase_modes.append(dict(mode=value, description=""))

        # Check: at least one element
        if len(phase_modes) < 2:
            issues.append((3, "Phase '"+current_phase+"' should have at least one mode"))

        # Check: no repetitions
        if len(phase_modes) != len(set([mode["mode"] for mode in phase_modes])):
            if current_phase.lower() != "code":
                issues.append((3, "There is at least a repeated mode in phase '"+current_phase+"'"))
            else:
                issues.append((3, "There is at least a repeated code in the list of codes"))

        # Update max column length
        if len(phase_modes) > max_len:
            max_len = len(phase_modes)

        if current_phase.lower() != "code":
            phases.append(phase_modes)
        else:
            codes = phase_modes[1:]

    # If not codes
    if not codes:
        codes = [str(i) for i in range(max_len-2, -1, -1)]

    return issues, None, dict(name=name, codes=codes, phases=phases)





import copy
import json
import logging
import os
import re
import tempfile
from io import BytesIO, StringIO
from typing import List, Union, BinaryIO, Tuple
from zipfile import ZipFile
import fire
import platform
import sys

print(f"Executable: {sys.executable}; V: {sys.version}; {platform.python_build()}; CONDA_PREFIX: {os.getenv('CONDA_PREFIX')}")

from nexinfosys.command_generators import Issue, IType
from nexinfosys.common.helper import download_file, generate_json
from nexinfosys.embedded_nis import NIS
from nexinfosys.model_services import State
from nexinfosys.serialization import deserialize_state


def set_zip_timestamp(in_zip: Union[os.PathLike, str, BinaryIO], timestamp=(2020, 1, 1, 0, 0, 0)) -> BinaryIO:
    """
    Modify the timestamp of all files, to stabilize the hash

    :param in_zip: Zip file whose files timestamp will be modified
    :param timestamp: Tuple with the timestamp to set for all the files
    :return A BytesIO with the resulting Zip
    """
    out_zip = BytesIO()
    with ZipFile(in_zip, mode="r") as zin:
        with ZipFile(out_zip, mode="w") as zout:
            for zinfo in zin.infolist():
                data = zin.read(zinfo.filename)
                zinfo_new = copy.copy(zinfo)
                zinfo_new.date_time = timestamp
                zout.writestr(zinfo_new, data)
    return out_zip


def hash_array(f: Union[str, bytes]):
    import hashlib
    m = hashlib.md5()
    if isinstance(f, str):
        m.update(f.encode("utf-8"))
    else:
        m.update(f)
    return m.digest()


def get_valid_name(original_name):
    prefix = original_name[0] if original_name[0].isalpha() else "_"
    remainder = original_name[1:] if original_name[0].isalpha() else original_name
    return prefix + re.sub("[^0-9a-zA-Z_]+", "", remainder)


def read_submit_solve_nis_file(nis_file_url: str, state: State, solve=False) -> Tuple[NIS, List[Issue]]:
    nis = NIS()
    nis.open_session(True, state)
    nis.reset_commands()
    # Load file and execute it
    nis.load_workbook(nis_file_url)
    if solve:
        issues = nis.submit_and_solve()
    else:
        issues = nis.submit()
    # tmp = nis.query_available_datasets()
    # print(tmp)
    # d = nis.get_results([(tmp[0]["type"], tmp[0]["name"])])
    return nis, issues


def issue_str(issue: Issue):
    # description, ctype, location
    return str(issue)


def prepare_base_state(base_url: str, solve: bool, directory: str = None, force_refresh: bool = False):
    from nexinfosys import initialize_configuration
    initialize_configuration()  # Needed to make download and NIS later work properly
    additional_slash = "/" if platform.system().lower() == "windows" else ""
    if base_url.startswith(os.sep) or base_url[1] == ":":
        base_url = f"file://{additional_slash}{base_url}"
    logging.debug(f"File to be downloaded: {base_url}")
    tmp_io = download_file(base_url)
    bytes_io = set_zip_timestamp(tmp_io)
    hash_ = hash_array(bytes_io.getvalue())
    val_name = get_valid_name(base_url)
    if directory is None:
        directory = tempfile.gettempdir()
    hash_file = f"{directory}{os.sep}base.hash.{val_name}"
    state_file = f"{directory}{os.sep}base.state.{val_name}"
    if force_refresh:
        update = True
    else:
        if os.path.isfile(hash_file) and os.path.isfile(state_file):
            with open(hash_file, "rb") as f:
                cached_hash = f.read()
            update = cached_hash != hash_
        else:
            update = True

    if update:
        temp_name = tempfile.NamedTemporaryFile(dir=directory, delete=False)
        temp_name = temp_name.name
        with open(temp_name, "wb") as f:
            f.write(bytes_io.getvalue())
        f_name = f"file://{additional_slash}{temp_name}"
        logging.debug(f"Temporary being loaded: {f_name}")
        nis, issues = read_submit_solve_nis_file(f_name, None, solve=solve)
        os.remove(temp_name)
        any_error = False
        for issue in issues:
            if issue.itype == IType.ERROR:
                any_error = True
        # Write if there are no errors
        if not any_error:
            # raise Exception(f'There were errors with the NIS base file {base_url}.')
            from nexinfosys.serialization import serialize_state
            state = nis.get_state()
            serial_state = serialize_state(state)
            nis.close_session()
            with open(hash_file, "wb") as f:
                f.write(hash_)
            with open(state_file, "wb") as f:
                f.write(serial_state)
    else:
        issues = None  # Cached, no issues
        with open(state_file, "rb") as f:
            serial_state = f.read()
        state = deserialize_state(serial_state)

    return state, serial_state, issues


def print_issues(operation, file, issues: List[Issue]):
    if issues is not None:
        if len(issues) > 0:
            error_count = 0
            warning_count = 0
            for issue in issues:
                if issue.itype == IType.ERROR:
                    error_count += 1
            if error_count > 0:
                print(f"There was a total of {error_count} errors detected")
            else:
                print(f"{operation} successful (no errors detected)")
            if warning_count > 0:
                print(f"There was a total of {warning_count} warnings")
            print("--------------------")
            for issue_type in [IType.ERROR, IType.WARNING]:
                if issue_type == IType.ERROR:
                    print(f"Errors ({error_count} total):")
                elif issue_type == IType.WARNING:
                    print(f"Warnings ({warning_count} total):")
                for issue in issues:
                    if issue.itype == issue_type:
                        print(issue_str(issue))
        else:
            print(f"{operation} successful, and no issues!")
    else:
        print(
            f"File '{file}' was already cached, meaning no change and no errors detected. "
            f"Please, add --force_refresh=True to force reprocessing.")


def write_results(state_or_nis: Union[State, NIS], output_dir: str, datasets: List[str]):
    if isinstance(state_or_nis, State):
        nis = NIS()
        nis.open_session(state=state_or_nis)
    else:  # NIS
        nis = state_or_nis

    # Convert to required format
    datasets = datasets.split(",") if datasets.strip() != "" else []
    _ = []
    for ds in datasets:
        tmp = ds.strip().split(".")
        if len(tmp) == 3:
            _.append([tmp[0], tmp[1]+"."+tmp[2]])
        elif len(tmp) == 2:
            _.append([None, tmp[0] + "." + tmp[1]])
        else:
            print(f"Could not parse dataset request entry: {ds}")
    tmp = nis.get_results(_)
    any_error = False
    for i, ds in enumerate(tmp):
        name = datasets[i]
        if ds[2]:
            if isinstance(ds[0], StringIO):
                val = ds[0].getvalue()
            else:
                val = ds[0]
            flag = "b" if not isinstance(ds[0], StringIO) and not isinstance(ds[0], str) else "t"
            print(f"Writing {name} to: {output_dir}{os.sep}{name}")
            with open(f"{output_dir}{os.sep}{name}", f"w{flag}") as f:
                f.write(val)
        else:
            any_error = True
            print(f"Could not retrieve dataset: {name}")
    if any_error or len(tmp) == 0:
        print("Reference of available datasets: --------")
        lst = nis.query_available_datasets()
        for ds in lst:
            fmts = '|'.join([fmt['format'] for fmt in ds['formats']])
            print(f"[{ds['type']}.]{ds['name']}.({fmts}); ({ds['description']})")


def set_log_level_from_cli_param(log_param: str):
    """
    Set global logging level to one of: Off, Info, Warn, Debug

    :param log_param:
    :return:
    """
    if log_param is None or log_param == "":
        log_param = ""

    level = log_param.lower()
    if level in ("debug", "d"):
        level = logging.DEBUG
    elif level == "off":
        level = logging.NOTSET
    elif level in ("error", "err", "e"):
        level = logging.DEBUG
    elif level in ("info", "i"):
        level = logging.INFO
    elif level in ("warn", "warning", "w"):
        level = logging.WARNING
    elif level in ("critical", "fatal"):
        level = logging.CRITICAL
    else:
        if level != "":
            print(f"Log level not set, '{log_param}' not recognized. Valid options: Error (E, Err), Debug (D), "
                  f"Warning (W, Warn), Info (I), Off, Critical (Fatal)")
        level = None

    if level is not None:
        logging.basicConfig(level=level)


class Nexinfosys:
    """
nexinfosys parse https://docs.google.com/spreadsheets/d/1BlvHI56kP2kzogKbAa8BHr2GYMfJIZTDkBaPYIfiPMQ/edit#gid=0 /home/rnebot/tmp/example1

nexinfosys solve https://docs.google.com/spreadsheets/d/1C5xNGvdORWqrWL6Ux2nBXMdX51ktEDJCmFdDrbFrhX4/edit#gid=0 /home/rnebot/tmp/example1withValues

nexinfosys solve https://docs.google.com/spreadsheets/d/1C5xNGvdORWqrWL6Ux2nBXMdX51ktEDJCmFdDrbFrhX4/edit#gid=0 /home/rnebot/tmp/example1withValues --datasets="flow_graph_solution.csv"

    """
    def parse(self, file: str, work_path: str, datasets: str = "", force_refresh: bool = False, log: str = None):
        """
        Parse and retrieve datasets

        :param file: Input file
        :param work_path: Where to place cache and result files
        :param datasets: Comma separated list of datasets to export
        :param force_refresh: True to force parsing even if the main file has not changed
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        os.makedirs(work_path, exist_ok=True)
        state, _, issues = prepare_base_state(file, False, work_path, force_refresh)
        print_issues("Parsing", file, issues)
        write_results(state, work_path, datasets)

    def solve(self, file: str, work_path: str, datasets: str = "", force_refresh: bool = False, log: str = None):
        """
        Parse, solve and retrieve datasets

        :param file: Input file
        :param work_path: Where to place cache and result files
        :param datasets: Comma separated list of datasets to export
        :param force_refresh: True to force parsing and solving even if the main file has not changed
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        os.makedirs(work_path, exist_ok=True)
        state, _, issues = prepare_base_state(file, True, work_path, force_refresh)
        print_issues("Solving", file, issues)
        write_results(state, work_path, datasets)

    def solve_dynamic_scenario(self, file: str, work_path: str, params_file_path: str = "",
                               datasets: str = "", force_refresh: bool = False, log: str = None):
        """
        Run (solve) dynamic scenario and get output datasets

        :param file: Input file
        :param work_path: Where to place cache and result files
        :param params_file_path: Path of the JSON file where parameters are specified. If empty, a sample params file is output to screen
        :param datasets: Comma separated list of datasets to export
        :param force_refresh: True to force parsing and solving even if the main file has not changed
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        os.makedirs(work_path, exist_ok=True)
        state, _, issues = prepare_base_state(file, True, work_path, force_refresh)
        print_issues("Dynamic scenario", file, issues)
        nis = NIS()
        nis.open_session(state=state)
        if params_file_path is None or params_file_path.strip() == "":
            # Print sample parameters file
            params = nis.query_parameters()
            d = {}
            for param in params:
                d[param["name"]] = f'{param["type"]}{": " if param["range"].strip() != "" else ""}{param["range"]}'
            print(generate_json(d))

        else:
            # Read file
            with open(params_file_path, "rt") as f:
                params = json.loads(f.read())
                issues = nis.recalculate(params)
                write_results(nis, work_path, datasets)


def main():
    os.environ["PAGER"] = "cat" if platform.system().lower() != "windows" else "-"
    fire.Fire(Nexinfosys)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
This package contains core models. It is divided in modules, a module per model topic. Some of the topics are
persistent (relational database) some are for memory (and they are persisted by serialization):

* Hierarchies in database
* Hierarchies in memory
* Statistical datasets
* MuSIASEM core concepts
* MuSIASEM generative concepts
* MuSIASEM problem specification
* MuSIASEM methodology support

"""
import logging
from collections import namedtuple

MODEL_VERSION = 1

# Ad-hoc structures
# CodeMutable = namedlist("Code", ["code", "description", "level", "children", "parents"])
# ConceptImmutable = namedtuple("ConceptTuple", "name istime description attributes code_list")
CodeImmutable = namedtuple("CodeTuple", "code description level children")
log_level = logging.DEBUG


from typing import Dict, List, Tuple, Set, Optional, Container
import networkx as nx

from nexinfosys.common.decorators import memoized_method
from nexinfosys.solving.graph import Node, EdgeType, Weight, Value


class ComputationGraph:
    """
    A Computation Graph is a directed graph where:
     - Each node represents a variable to be computed or which value is given.
     - Each directed edge represents how the destination node can be computed based on the value of the source node, or
       in the opposite direction if a reverse weight exist.
     - Computations can be done in 'direct' order (Top-Down strategy) using the direct weight, and in the 'reverse'
       order (Bottom-Up strategy) using the reverse weight.

    How node values are calculated?

     - General rule: Summing up the weighted input values. This is, the selected node value is the sum of each incoming
                     (predecessor) node value multiplied by the "weight" attribute associated with the edge from the
                     incoming node.
     - Split rule: If the node value is split entirely (100%) into the successors nodes, its value can be computed in
                   the opposite direction (from successors) only by computing one of the input nodes.
    """
    def __init__(self, graph: Optional[nx.DiGraph] = None, name: str = None):
        self.graph = nx.DiGraph()
        self.name = name

        self.descendants: Optional[Dict[Node, Set[Node]]] = None

        if graph:
            for u, v, data in graph.edges(data=True):
                reverse_weight = (Weight(1.0)/data["weight"]) if data.get("add_reverse_weight") else None
                self.add_edge(u, v, data["weight"], reverse_weight)

            for n, data in graph.nodes(data=True):
                if data.get("add_split"):
                    self.mark_node_split(n, EdgeType.DIRECT)

            self.compute_descendants()

    @property
    def nodes(self):
        """ Return the nodes of the flow graph """
        return self.graph.nodes

    def nodes_not_in_container(self, container: Container[Node]) -> List[Node]:
        """ Return the nodes of the graph not present in a container """
        return [n for n in self.graph.nodes if n not in container]

    def nodes_in_container(self, container: Container[Node]) -> List[Node]:
        """ Return the nodes of the graph present in a container """
        return [n for n in self.graph.nodes if n in container]

    @memoized_method(maxsize=None)
    def _inputs(self, n: Node, edge_type: EdgeType) -> List[Tuple[Node, Optional[Weight]]]:
        """ Return the predecessors of a node in the specified direction """
        return [(e[0], e[2]['weight']) for e in self.graph.in_edges(n, data=True) if e[2]['type'] == edge_type]

    def direct_inputs(self, n: Node):
        """ Return the predecessors of a node in the 'direct' (Top-Down) direction """
        return self._inputs(n, EdgeType.DIRECT)

    def reverse_inputs(self, n: Node):
        """ Return the predecessors of a node in the 'reverse' (Bottom-Up) direction """
        return self._inputs(n, EdgeType.REVERSE)

    def weighted_successors(self, n: Node) -> List[Tuple[Node, Weight]]:
        """ Return the successors of a node that have a valid weight """
        return [suc for suc in self.graph.successors(n) if self.graph[n][suc]['weight'] is not None]

    def add_edge(self, u: Node, v: Node, weight: Optional[Weight], reverse_weight: Optional[Weight]):
        """ Add an edge with weight attributes to the computation graph """
        self.graph.add_edge(u, v, weight=weight, type=EdgeType.DIRECT)
        self.graph.add_edge(v, u, weight=reverse_weight, type=EdgeType.REVERSE)
        self.init_node_split(u)
        self.init_node_split(v)

        # Invalidate descendants, it must be recomputed
        self.descendants = None

    def init_node_split(self, n: Node) -> None:
        """ Set the default value for attribute 'split' to a node """
        if not self.graph.nodes[n].get("split"):
            self.graph.nodes[n]["split"] = [False, False]

    def mark_node_split(self, n: Node, graph_type: EdgeType, split: bool = True) -> None:
        """ Set the attribute 'split' to a node """
        self.graph.nodes[n]["split"][graph_type.value] = split

    def get_node_split(self, n: Node, graph_type: EdgeType) -> bool:
        """ Get the attribute 'split' from a node """
        return self.graph.nodes[n]["split"][graph_type.value]

    def get_direct_node_split(self, n: Node) -> bool:
        """ Get the DIRECT attribute 'split' from a node """
        return self.get_node_split(n, EdgeType.DIRECT)

    def get_reverse_node_split(self, n: Node) -> bool:
        """ Get the REVERSE attribute 'split' from a node """
        return self.get_node_split(n, EdgeType.REVERSE)

    def compute_descendants(self):
        def visit_forward(node: Node) -> None:
            visited_nodes.add(node)

            if node in self.descendants:
                visited_nodes.update(self.descendants[node])
                return

            for successor in self.weighted_successors(node):
                if successor not in visited_nodes:
                    visit_forward(successor)

        self.descendants = {}

        for n in self.graph.nodes:  # type: Node
            visited_nodes: Set[Node] = set()
            visit_forward(n)
            self.descendants[n] = visited_nodes - {n}

    def compute_conflicts(self, new_computed_nodes: Set[Node], prev_computed_nodes: Set[Node]) -> Dict[Node, Set[Node]]:
        _new_computed_nodes = self.nodes_in_container(new_computed_nodes)
        if not _new_computed_nodes:
            return {}  # Nothing new to compute

        if not self.descendants:
            self.compute_descendants()

        _prev_computed_nodes = self.nodes_in_container(prev_computed_nodes)
        return self._compute_conflicts(set(_new_computed_nodes), set(_prev_computed_nodes))

    def _compute_conflicts(self, new_computed_nodes: Set[Node], prev_computed_nodes: Set[Node]) -> Dict[Node, Set[Node]]:
        assert self.descendants
        conflicts: Dict[Node, Set[Node]] = {}
        computed_nodes = prev_computed_nodes | new_computed_nodes

        for node in prev_computed_nodes:
            intersection = self.descendants[node] & new_computed_nodes
            if intersection:
                conflicts[node] = intersection

        for node in new_computed_nodes:
            intersection = self.descendants[node] & computed_nodes
            if intersection:
                conflicts[node] = intersection

        return conflicts

    def compute_param_conflicts(self, params: Set[Node]) -> Dict[Node, Set[Node]]:
        """ Calculate the conflicts between nodes with values - the parameters - in a computation graph.
            If node A has a conflict with node B, means that B can be entirely or partially computed from node A,
            so both nodes cannot have input values at the same time (unless these values are consistent).

            Example]
            - Given the graph: a -> b -> c -> e <- d
            - Given the parameters: a, b, c, d
            - The conflicts are:
                a: {b, c}  (b and c can be computed from a)
                b: {c} (c can be computed from b)
                c: none
                d: none

            **DEPRECATED METHOD**

            :param params: the set of parameters of a computation graph, i.e. the nodes that have values and we use
                           to compute the values in the remaining nodes of the graph.
            :return: a dictionary with an entry for each parameter where the value is a set with the names of other
                     conflicting parameters
        """
        def visit_forward(node: Node) -> Set[Node]:
            visited_nodes.add(node)

            result = set()

            if node in sub_params:
                result.add(node)

            for suc in self.weighted_successors(node):
                if suc not in visited_nodes:
                    result |= visit_forward(suc)

            return result

        all_conflicts: Dict[Node, Set[Node]] = {}
        # Filter out params not in nodes
        filtered_params = {n for n in params if n in self.nodes}

        for param in filtered_params:
            sub_params: Set[Node] = filtered_params - {param}
            visited_nodes: Set[Node] = set()
            conflicts = visit_forward(param)
            all_conflicts[param] = conflicts

        return all_conflicts

    @staticmethod
    def compute_param_combinations(conflicts: Dict[Node, Set[Node]]) -> Set[frozenset]:
        """ Given the conflicts between parameters in a graph, compute the valid combinations of them.
            This is, if there are different nodes that can have a value, compute which combination of them can be used
            when calculating the values in the rest of nodes without having a conflict.
            A "conflict" should be read as: different ways of computing the value for a node.

            Example
            - Given the graph: a -> b -> c -> e <- d
            - Given the parameters: a, b, c, d
            - The conflicts are:
                a: {b, c}  (b and c can be computed from a)
                b: {c} (c can be computed from b)
                c: none
                d: none
            - The valid combinations are:
                {a, d}
                {b, d}
                {c, d}

            :param conflicts: a dictionary with an entry for each parameter where the value is a set with the names
                              of other conflicting parameters
            :return: a set with different combinations (sets) of parameters that can be used
        """
        def valid_combinations(param: Node, params: Set[Node]) -> Set[frozenset]:
            result: Set[frozenset] = set()
            other_params: Set[Node] = params - conflicts[param] - {param}
            non_conflicting = set()

            for other in other_params:
                if param not in conflicts[other]:
                    non_conflicting.add(other)

            for other in non_conflicting:
                for comb in valid_combinations(other, non_conflicting):
                    result |= {comb | frozenset({param})}

            if len(result) == 0:
                result = {frozenset({param})}

            return result

        combinations: Set[frozenset] = set()
        param_names = set(conflicts.keys())

        # This is a shortcut to avoid computation
        if all([len(s) == 0 for s in conflicts.values()]):
            return {frozenset(param_names)}

        # TODO: This routine needs optimization because its order of complexity is too big.
        #  It takes lot of time even with 10 parameters!

        for p in param_names:
            combinations |= valid_combinations(p, param_names)

        return combinations

    def compute_values(self, nodes: List[Node], params: Dict[Node, Value]) \
            -> Tuple[Dict[Node, Optional[Value]], Dict[Node, Optional[Value]]]:
        """ Given a computation graph and a set of nodes with values (the parameters) compute the values of
            a list of nodes.

        :param nodes: the list of nodes whose value we are interested in
        :param params: a dictionary with an entry for each parameter and its value
        :return: a tuple with 1) the computed values for the desired nodes 2) computed values for other nodes during
                 the process.
        """
        def solve_inputs(inputs: List[Tuple[Node, Weight]], split: bool) -> Optional[Value]:
            input_values: List[Tuple[Value, Weight]] = []

            for n, weight in sorted(inputs):
                res_backward = solve_backward(n)

                # If node 'n' is a 'split' only one result is needed to compute the result
                if split:
                    if res_backward is not None:
                        return res_backward * weight
                else:
                    if res_backward is not None and weight is not None:
                        input_values.append((res_backward, weight))
                    else:
                        return None

            return Value.compute_weighted_addition(input_values)

        def solve_backward(node: Node) -> Optional[Value]:
            # Is the node already computed?
            if node in values:
                return values[node]

            # Does a parameter exist for this node?
            param = params.get(node)
            if param is not None:
                values[node] = param
                return param

            if node in pending_nodes:
                return None

            pending_nodes.append(node)

            split = self.graph.nodes[node]["split"]

            result = solve_inputs(self.direct_inputs(node), split[EdgeType.REVERSE.value])

            if result is None:
                result = solve_inputs(self.reverse_inputs(node), split[EdgeType.DIRECT.value])

            if result is not None:
                # TODO: this line is not generic
                result.name = node.name

            values[node] = result
            return result

        results: Dict[Node, Optional[Value]] = {}
        values: Dict[Node, Optional[Value]] = {}

        for n in nodes:
            pending_nodes: List[Node] = []
            value = solve_backward(n)
            results[n] = value

        return results, values



# -*- coding: utf-8 -*-
import pandas as pd
from typing import List

from nexinfosys.models.statistical_datasets import DataSource, Database, Dataset
from nexinfosys.ie_imports.data_source_manager import IDataSourceManager


def get_ssp_dimension_names_dataset(dataset_name):
    if dataset_name.lower()=="regions":
        return ["source_id", "model", "scenario", "spatial", "temporal", "variable"]
    elif dataset_name.lower()=="countries":
        return []


def get_ssp_dataset(dataset_name: str, convert_dimensions_to_lower=False):
    """
    Read into a Dataframe the requested SSP dataset

    :param dataset_name: either "regions" or "countries"
    :param convert_dimensions_to_lower: True is to convert dimensions to lower case
    :return: A DataFrame with the dataset
    """
    # Read some configuration knowing where the ssp is stored
    if "SSP_FILES_DIR" in app.config:
        base_path = app.config["SSP_FILES_DIR"]
    else:
        base_path = "/home/rnebot/GoogleDrive/AA_MAGIC/Data/SSP/"
    fname = base_path + "/" + dataset_name.lower() + ".csv"
    # Read the file into a Dataframe
    df = pd.read_csv(fname)
    dims = get_ssp_dimension_names_dataset(dataset_name)
    # Convert to lower case if needed
    if convert_dimensions_to_lower:
        for d in dims:
            df[d] = df[d].astype(str).str.lower()
    # Index the dataframe on the dimensions
    df.set_index(dims, inplace=True)

    # Return the Dataframe
    return df


class SSPLocal(IDataSourceManager):
    def __init__(self):
        pass

    def get_name(self) -> str:
        """ Source name """
        pass

    def get_datasource(self) -> DataSource:
        """ Data source """
        return "SSP"

    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        return ["SSP"]

    def get_datasets(self, database=None) -> list:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """
        return {"regions": "SSP by regions (6) and global",
                "countries": "SSP by country"
                }

    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """
        s = get_ssp_dimension_names_dataset(dataset)
        pass

    def etl_full_database(self, database=None, update=False):
        """ If bulk download is supported, refresh full database """
        pass

    def etl_dataset(self, dataset, update=False):
        """ If bulk download is supported, refresh full dataset """
        pass

    def get_dataset_filtered(self, dataset, dataset_params: list) -> Dataset:
        """ Obtains the dataset with its structure plus the filtered values
            The values can be in a pd.DataFrame or in JSONStat compact format
            After this, new dimensions can be joined, aggregations need to be performed
        """
        pass

    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass



# -*- coding: utf-8 -*-

from nexinfosys.models.musiasem_concepts import *


class Variable:
    """ Used by ExpressionsEvaluationEngine to represent variables or parameters """
    def __init__(self, name):
        self._name = name
        self._type = None  # Parameter, factor, taxon, factortype, indicator
        self._origin = None  # The original object (Parameter, Factor, Taxon, FactorType, Indicator)
        self._values = []  # List of tuples (Literal, Expression). Expression refers to the one that caused obtaining the value, if None it is a "given", using
        # List of expressions where the variable appears. Each element is a tuple (expression, left|right).
        # "left"  if the variable appears to the left
        # "right" if the variable appears to the right
        self._expressions = []

    @property
    def values(self):
        return self._values

    def values_append(self, f: float, e: "Expression"=None):
        self._values.append((f, e))

    def values_set(self, f: float, e: "Expression"=None):
        self._values = [(f, e)]

    @property
    def expressions(self):
        return self._expressions

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, o):
        self._origin = o


class Expression:
    """ Used by ExpressionsEvaluationEngine to represent expressions """
    def __init__(self, e: str, expression: dict):
        self._expression = expression
        if isinstance(e, str):
            self._expression_string = e
        else:
            self._expression_string = None
        self._origin = None  # The original object (QualifiedQuantityExpression, Factor, FactorType, Taxon, TODO: Mapping)
        self._vars = None  # Variables mentioned in the expression

    @property
    def expression(self):
        return self._expression

    @property
    def origin(self):
        return self._origin

    @property
    def vars(self):
        return self._vars

    @vars.setter
    def vars(self, v):
        self._vars = v

    @origin.setter
    def origin(self, o):
        self._origin = o

    # ############################################################################################################ #
    @staticmethod
    def expressions_from_entity(entity, workspace=None) -> Union[List[str], List[dict]]:
        """ Elaborates a list of expressions in string or direct dictionary format """
        def get_factor_name(f: Factor):
            return f.processor.name+"."+f.name

        def add_taxon_hierarchy(h: Hierarchy) -> List[Expression]:
            """ Traverse the hierarchy and generate expressions. The expression is always:
                parent node = SUM children nodes. The sum may be weighted. The expression could be also given for the node
                if the node is a leaf node. The case for a parent node with a given expression should be studied, if two
                expressions should be generated or if
            """
            return []

        lst = []
        if isinstance(entity, FactorQuantitativeObservation):
            lst.append( get_factor_name(entity.factor) + "=" + entity.expression )
        elif isinstance(entity, Hierarchy):
            # TODO Find connections from and connections to
            # TODO Check that the hierarchy is of Taxon or FactorType
            lst = add_taxon_hierarchy(entity)
        elif isinstance(entity, Indicator):
            # TODO This will probably require functions to select a variable number of processors
            pass
        elif isinstance(entity, dict): # Direct expression
            lst.append(entity)

        return lst


class ExpressionsEngine:
    """ The procedure to evaluate is:
        1) Call one or more times "add_expression"
          Internally it calls "parse" and "find_vars"
          Prepare a dependency graph. Both vars and expressions are nodes, undirected arcs mean that a node conditions
          or is conditioned by other
        2) Call one or more times "add_variable_value"
          Appends
        3) Call "solve"
          Prepare a set pointing to ALL the expressions
          For each expression in the set,
            Call "eval"
            If the result is a variable with a solution, remove the expression from the list and 
              process the expressions in which the variable appears  
           
        4) "reset" to use the same expressions
           "change_variable_value" to 
    """
    def __init__(self, workspace=None):
        self._workspace = workspace  # To retrieve variables by their name
        self._variables = create_dictionary()  # type: Dictionary
        self._expressions = []
        # Must be set before internally calling "_eval"
        self._current_evaluated_expression = None  # type: Expression

    def reset(self):
        self.reset_expressions()
        self.reset_variables()

    def reset_expressions(self):
        self._expressions = []

    def reset_variables(self):
        self._variables = create_dictionary()

    def reset_solution_values(self):
        for k, v in self._variables.items():
            # If the value came from direct input, keep it
            # If the value came from the evaluation of expressions, remove it
            v._values = [val for val in v._values if not val[1]]

    def append_expressions(self, origin):
        """ Create the expression, add the appropriate records, and analyze it """
        # First, convert the object to a list of expression strings
        lst = Expression.expressions_from_entity(origin)
        for e in lst:
            # Parse the expression
            e_json = ExpressionsEngine._parse(e)
            # Create the Expression object
            expression = Expression(e, e_json)
            # Extract the variables appearing in the expression
            vars = []
            for v in ExpressionsEngine._find_vars(e_json):
                if v not in self._variables:
                    va = Variable(v)
                    va._origin = expression.origin
                    self._variables[v] = va
                else:
                    va = self._variables[v]
                # Make the variable point to the expression
                va.expressions.append(expression)
                # Make the expression point to the variable
                vars.append(va)
            # Add Expression to the list of expressions
            expression.origin = origin
            expression.vars = vars
            self._expressions.append(expression)

    def append_variable_value(self, n: str, f: float):
        if n in self._variables:
            self._variables[n].values_append(f)  # Given value. No Expression
        else:
            raise Exception("Variable '"+n+"' not found when trying to append a value to it")

    def set_variable_value(self, n: str, f: float):
        if n in self._variables:
            self._variables[n] = [(f, None)]
        else:
            raise Exception("Variable '"+n+"' not found when trying to set its value")

    @property
    def variables(self):
        return self._variables

    @staticmethod
    def _parse(exp_str):
        """ This needs the definition of a syntax, generation of syntactic and lexical rules, use of a component
            considering these rules to generate the JSON used by EVAL
        """
        if isinstance(exp_str, dict):
            j = exp_str  # Assume it is already a dictionary
        else:
            try:
                j = json.loads(exp_str)
            except Exception:
                # TODO Parse normally... "None" if some error ocurred (and an exception pointing to the problem)
                j = None

        return j

    @staticmethod
    def _find_vars(exp_json):
        """ Scan the expression to find variables, so they can be indexed """

        def find_vars_recurr(curr_json):
            if "lhs" in exp_json and "rhs" in curr_json:  # Equation
                find_vars_recurr(exp_json["lhs"])
                find_vars_recurr(exp_json["rhs"])
            elif "op" in curr_json and "oper" in curr_json:  # Arithmetic operations
                for i in curr_json["oper"]:
                    find_vars_recurr(i)
            elif "v" in curr_json:  # A variable
                # Take variable name
                # TODO Maybe standardize variable name. The same for evaluation
                vars_set.add(curr_json["v"])

        vars_set = set()
        find_vars_recurr(exp_json)
        return vars_set

    def _eval(self, exp_json):
        """ Evaluate a dictionary JSON containing the already parsed expression.
            It can return a value if all variables are known (needs access to a registry),
            True or False if it is an equation with no unknowns (it would be a comparison),
            a Variable if the expression solves to the value of Variable, or None if there is more than one unknown """
        if "lhs" in exp_json and "rhs" in exp_json:  # Equation
            lhs = self._eval(exp_json["lhs"])
            rhs = self._eval(exp_json["rhs"])
            n = (1 if isinstance(lhs, Variable) else 0) + (1 if isinstance(rhs, Variable) else 0)
            if n == 1:  # Only one Variable
                if isinstance(lhs, Variable):
                    v = lhs
                    q = rhs
                else:
                    v = rhs
                    q = lhs
                if q:
                    if isinstance(q, (int, float, ureg.Quantity)):
                        # Assign q to the Variable "v"
                        v.values_append(q, self._current_evaluated_expression)
                        return v  # <<<<<<<<<
                    else:
                        raise Exception("Unsupported type '"+str(type(q)))
                else:  # If the result is None, the Variable cannot be solved, the Expression is not evaluable
                    return None
            elif n == 2:  # Two variables
                # Return None, because both variables have no value
                return None
            else: # 0
                if type(lhs) == type(rhs):
                    return lhs == rhs
                else:
                    raise Exception("Incorrect expression. ??: "+str(self._current_evaluated_expression.expression))

        elif "op" in exp_json and "oper" in exp_json:  # Arithmetic operations
            # TODO When many observations appear, here is where combinatorial explosion can appear
            # TODO For instance, instead of just multiplying single multiplicands, assume each
            # TODO multiplicand can be a list of numbers. "itertools.product" could be used for this
            r_lst = [self._eval(i) for i in exp_json["oper"]]
            op = exp_json["op"]
            if op in ("*", "/"):
                acum = 1
            elif op in ("+", "-"):
                acum = 0

            for r in r_lst:
                if r:
                    if isinstance(r, (int, float, ureg.Quantity)):
                        if op == "*":
                            acum *= r
                        elif op == "/":
                            acum /= r
                        elif op == "+":
                            acum += r
                        elif op == "-":
                            acum -= r
                    elif isinstance(r, Variable): # Having a Variable here means we have an unknown, which is a stopper for the cascade solver
                        return None
                    else: # TODO Not controlled conditions also return None
                        return None
                else:  # TODO Not controlled conditions, return None
                    return None
            return acum
        elif "v" in exp_json:  # A variable. It can be anything: factor observation, parameter, ¿indicator?
            # Obtain the variable object and check if the variable has a value
            v = self._variables[exp_json["v"]]
            if v:
                if len(v.values) > 0:
                    return v.values[0][0]  # TODO Only one solution for now. Consider how to process several solutions. ¿"itertools.product"?
                else:
                    return v
            else:
                raise Exception("Variable '"+exp_json("v")+"' not found. What happened?")
        elif "n" in exp_json:  # A quantity
            # Because we are evaluating numerically, for now use only the number and the unit
            # TODO In the future, the uncertainty could also be used
            q = exp_json["n"]
            if "u" in exp_json:
                # Check validity of the unit with Pint
                try:
                    u = ureg(exp_json["u"])
                except pint.errors.UndefinedUnitError:
                    # The user should know that the specified unit is not recognized
                    raise Exception("The specified unit '" + exp_json["u"] + "' is not recognized")
                q *= u
            else:
                q *= ureg("dimensionless")
            return q
        elif "f" in exp_json:  # A function call. A world of options...
            return None

    @staticmethod
    def generate_expression(exp_json):
        """ Generate an expression string from JSON. It involves generating proper variable names """
        pass

    @staticmethod
    def generate_expression_for_sympy(exp_json):
        # TODO Rename variables using Sympy valid names
        # TODO Keep a map of these renames, from Sympy to MuSIASEM name and also vice versa
        # TODO The first is needed to fill results back after solving
        # TODO The second to assign already known variable values
        # TODO When renaming try to keep as much as possible the original name, so the generated
        # TODO equations can be readable
        pass

    @staticmethod
    def is_right_side(exp_json):
        pass

    def cascade_solver(self):
        """ Simple solver based on cascading of already known variables.
          Prepare a set pointing to ALL the expressions
          For each expression in the set,
            Call "eval"
            If the result is a variable with a solution, remove the expression from the list and
              process the expressions in which the variable appears

        """
        def evaluate(e: Expression):
            nonlocal some_change
            self._current_evaluated_expression = e
            r = self._eval(e.expression)
            if r and isinstance(r, Variable):  # A Variable has value and the Expression is solved closed
                some_change = True
                # Do not evaluate it more
                open_expressions.remove(e)
                # Remove the expression from the variables pointing to it
                for v in e.vars:
                    v.expressions.remove(e)
                # Now, evaluate the expressions related to the new found Variable
                for e in r.expressions:
                    evaluate(e)
            if e in pending_expressions:
                pending_expressions.remove(e)

        # Set of expressions where things can change
        open_expressions = set(self._expressions)
        # Reset all values and states. Except parameters.
        self.reset_solution_values()
        # Sweep expressions and repeat while there are new
        some_change = True
        while some_change:
            some_change = False
            # Set of expressions inside open_expressions which have not been iterated
            pending_expressions = open_expressions.copy()
            while pending_expressions:
                # Take one expression from the set
                exp = pending_expressions.pop()
                pending_expressions.add(exp)
                # Evaluate it (and if the solution to a Variable is found, recurse to related Expressions)
                evaluate(exp)




######################
#  LIST OF COMMANDS  #
######################
from typing import List

from nexinfosys import Command, regex_var_name, regex_hvar_name, regex_cplex_var, CommandType
from nexinfosys.command_generators.spreadsheet_command_parsers.external_data.etl_external_dataset_spreadsheet_parse import \
    parse_etl_external_dataset_command
from nexinfosys.command_generators.spreadsheet_command_parsers.external_data.mapping_spreadsheet_parse import \
    parse_mapping_command
from nexinfosys.command_generators.spreadsheet_command_parsers.specification.data_input_spreadsheet_parse import \
    parse_data_input_command
from nexinfosys.command_generators.spreadsheet_command_parsers.specification.hierarchy_spreadsheet_parser import \
    parse_hierarchy_command
from nexinfosys.command_generators.spreadsheet_command_parsers.specification.metadata_spreadsheet_parse import \
    parse_metadata_command
from nexinfosys.command_generators.spreadsheet_command_parsers.specification.pedigree_matrix_spreadsheet_parse import \
    parse_pedigree_matrix_command
from nexinfosys.command_generators.spreadsheet_command_parsers.specification.references_spreadsheet_parser import \
    parse_references_command
from nexinfosys.command_generators.spreadsheet_command_parsers.specification.scale_conversion_spreadsheet_parse import \
    parse_scale_conversion_command
from nexinfosys.command_generators.spreadsheet_command_parsers.specification.structure_spreadsheet_parser import \
    parse_structure_command
from nexinfosys.command_generators.spreadsheet_command_parsers.specification.upscale_spreadsheet_parse import \
    parse_upscale_command
from nexinfosys.command_generators.spreadsheet_command_parsers_v2.dataset_data_spreadsheet_parse import \
    parse_dataset_data_command
from nexinfosys.command_generators.spreadsheet_command_parsers_v2.dataset_qry_spreadsheet_parse import \
    parse_dataset_qry_command

# BE CAREFUL: the order of the commands in the list MATTERS because it is used to prioritize the detection
# of the command based on the worksheet name and the allowed_names / alt_regex of the command.
# E.g. the command "cat_hier_mapping" allows a worksheet name with the pattern "HierarchiesMappings.*" and the command
# "cat_hierarchies" allows the pattern "Hierarchies.*". An actual worksheet named "HierarchiesMappings Spain" would
# match both patterns, so the command of the larger pattern should come first in the list below.
commands: List[Command] = [
    Command(name="dummy", allowed_names=["Dummy"], is_v1=True,
            cmd_type=CommandType.misc,
            execution_class_name="nexinfosys.command_executors.specification.dummy_command.DummyCommand"),

    # Version 1 and maybe also Version 2

    Command(name="references", allowed_names=["References"], is_v1=True, is_v2=True,
            cmd_type=CommandType.metadata,
            execution_class_name="nexinfosys.command_executors.specification.references_command.ReferencesCommand",
            parse_function=parse_references_command,
            alt_regex=r"(References|Ref)[ _]" + regex_var_name),

    # Version 1 and Version 2

    Command(name="metadata", allowed_names=["Metadata"], is_v1=True, is_v2=True,
            cmd_type=CommandType.metadata,
            execution_class_name="nexinfosys.command_executors.specification.metadata_command.MetadataCommand",
            parse_function=parse_metadata_command),

    Command(name="pedigree_matrix", allowed_names=["Pedigree"], is_v1=True, is_v2=True,
            cmd_type=CommandType.metadata,
            execution_class_name="nexinfosys.command_executors.specification.pedigree_matrix_command.PedigreeMatrixCommand",
            parse_function=parse_pedigree_matrix_command,
            alt_regex=r"(PedigreeV1|Ped|NUSAP\.PM)[ _]+" + regex_var_name),

    # Version 2 only

    Command(name="cat_hier_mapping",
            allowed_names=["CodeHierarchiesMapping", "HierarchiesMapping"],
            is_v2=True, cmd_type=CommandType.input,
            execution_class_name="nexinfosys.command_executors.version2.hierarchy_mapping_command.HierarchyMappingCommand"),

    Command(name="cat_hierarchies", allowed_names=["CodeHierarchies", "Hierarchies"],
            is_v2=True, cmd_type=CommandType.input,
            execution_class_name="nexinfosys.command_executors.version2.hierarchy_categories_command.HierarchyCategoriesCommand"),

    # Command(name="attribute_types", allowed_names=["AttributeTypes"], is_v2=True,
    #         cmd_type=CommandType.misc,
    #         execution_class_name="nexinfosys.command_executors.version2.attribute_types_command.AttributeTypesCommand"),

    Command(name="parameters", allowed_names=["Parameters", "Params"], is_v2=True,
            cmd_type=CommandType.input,
            execution_class_name="nexinfosys.command_executors.external_data.parameters_command.ParametersCommand"),

    Command(name="datasetdef", allowed_names=["DatasetDef"], is_v2=True,
            cmd_type=CommandType.input,
            execution_class_name="nexinfosys.command_executors.version2.dataset_definition_command.DatasetDefCommand"),

    # Command(name="attribute_sets", allowed_names=["AttributeSets"], is_v2=True,
    #         cmd_type=CommandType.misc,
    #         execution_class_name="nexinfosys.command_executors.version2.attribute_sets_command.AttributeSetsCommand"),

    Command(name="interface_types", allowed_names=["InterfaceTypes"], is_v2=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.version2.interface_types_command.InterfaceTypesCommand"),

    Command(name="processors", allowed_names=["BareProcessors"], is_v2=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.version2.processors_command.ProcessorsCommand"),

    Command(name="interfaces_and_qq", allowed_names=["Interfaces"], is_v2=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.version2.interfaces_command.InterfacesAndQualifiedQuantitiesCommand"),

    Command(name="relationships", allowed_names=["Relationships", "Relationship", "Flows"], is_v2=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.version2.relationships_command.RelationshipsCommand"),

    Command(name="processor_scalings", allowed_names=["ProcessorScalings", "ProcessorScaling"], is_v2=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.version2.processor_scalings_command.ProcessorScalingsCommand"),

    Command(name="scale_conversion_v2", allowed_names=["ScaleChangeMap", "ScaleChangeMaps"], is_v2=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.version2.scale_conversion_v2_command.ScaleConversionV2Command"),

    Command(name="problem_statement", allowed_names=["ProblemStatement"], is_v2=True,
            cmd_type=CommandType.analysis,
            execution_class_name="nexinfosys.command_executors.version2.problem_statement_command.ProblemStatementCommand"),

    Command(name="scalar_indicator_benchmarks", allowed_names=["ScalarBenchmarks"], is_v2=True,
            cmd_type=CommandType.analysis,
            execution_class_name="nexinfosys.command_executors.version2.scalar_indicator_benchmarks_command.ScalarIndicatorBenchmarksCommand"),

    Command(name="scalar_indicators", allowed_names=["ScalarIndicators"], is_v2=True,
            cmd_type=CommandType.analysis,
            execution_class_name="nexinfosys.command_executors.version2.scalar_indicators_command.ScalarIndicatorsCommand"),

    Command(name="matrix_indicators", allowed_names=["MatrixIndicators"], is_v2=True,
            cmd_type=CommandType.analysis,
            execution_class_name="nexinfosys.command_executors.version2.matrix_indicators_command.MatrixIndicatorsCommand"),

    Command(name="lcia_methods", allowed_names=["LCIAMethods"], is_v2=True,
            cmd_type=CommandType.analysis,
            execution_class_name="nexinfosys.command_executors.version2.lcia_methods_command.LCIAMethodsCommand"),

    Command(name="datasetdata", allowed_names=["DatasetData"], is_v2=True,
            cmd_type=CommandType.input,
            execution_class_name="nexinfosys.command_executors.version2.dataset_data_command.DatasetDataCommand",
            parse_function=parse_dataset_data_command),

    Command(name="datasetqry", allowed_names=["DatasetQry"], is_v2=True,
            cmd_type=CommandType.input,
            execution_class_name="nexinfosys.command_executors.version2.dataset_query_command.DatasetQryCommand",
            parse_function=parse_dataset_qry_command),

    Command(name="ref_pedigree_matrices", allowed_names=["PedigreeMatrices"], is_v2=True,
            cmd_type=CommandType.metadata,
            execution_class_name="nexinfosys.command_executors.version2.pedigree_matrices_command.PedigreeMatricesReferencesCommand"),

    Command(name="ref_provenance", allowed_names=["RefProvenance"], is_v2=True,
            cmd_type=CommandType.metadata,
            execution_class_name="nexinfosys.command_executors.version2.references_v2_command.ProvenanceReferencesCommand"),

    Command(name="ref_geographical", allowed_names=["RefGeographic", "RefGeography"], is_v2=True,
            cmd_type=CommandType.metadata,
            execution_class_name="nexinfosys.command_executors.version2.references_v2_command.GeographicReferencesCommand"),

    Command(name="ref_bibliographic", allowed_names=["RefBibliographic", "RefBibliography"], is_v2=True,
            cmd_type=CommandType.metadata,
            execution_class_name="nexinfosys.command_executors.version2.references_v2_command.BibliographicReferencesCommand"),

    Command(name="list_of_commands",
            allowed_names=["ListOfCommands"],
            is_v2=True, cmd_type=CommandType.convenience,
            execution_class_name=None),

    Command(name="import_commands",
            allowed_names=["ImportCommands"],
            is_v2=True, cmd_type=CommandType.convenience,
            execution_class_name=None),

    # Version 1 only

    Command(name="data_input", allowed_names=["DataInput"], is_v1=True,
            cmd_type=CommandType.input,
            execution_class_name="nexinfosys.command_executors.specification.data_input_command.DataInputCommand",
            parse_function=parse_data_input_command,
            alt_regex=r"(Processors|Proc)[ _]+" + regex_var_name),

    Command(name="hierarchy", allowed_names=["Hierarchy"], is_v1=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.specification.hierarchy_command.HierarchyCommand",
            parse_function=parse_hierarchy_command,
            alt_regex=r"(Taxonomy|Tax|Composition|Comp)[ _]([cpf])[ ]" + regex_var_name),

    Command(name="upscale", allowed_names=["Upscale"], is_v1=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.specification.upscale_command.UpscaleCommand",
            parse_function=parse_upscale_command,
            alt_regex=r"(Upscale|Up)[ _](" + regex_var_name + "[ _]" + regex_var_name + ")?"),

    Command(name="structure", allowed_names=["Structure"], is_v1=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.specification.structure_command.StructureCommand",
            parse_function=parse_structure_command,
            alt_regex=r"(Grammar|Structure)([ _]+" + regex_var_name + ")?"),

    Command(name="scale_conversion", allowed_names=["Scale"], is_v1=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.specification.scale_conversion_command.ScaleConversionCommand",
            parse_function=parse_scale_conversion_command,
            alt_regex=r"Scale"),

    Command(name="etl_dataset", allowed_names=["Dataset"], is_v1=True,
            cmd_type=CommandType.input,
            execution_class_name="nexinfosys.command_executors.external_data.etl_external_dataset_command.ETLExternalDatasetCommand",
            parse_function=parse_etl_external_dataset_command,
            alt_regex=r"(Dataset|DS)[ _]" + regex_hvar_name),

    Command(name="mapping", allowed_names=["Mapping"], is_v1=True,
            cmd_type=CommandType.core,
            execution_class_name="nexinfosys.command_executors.external_data.mapping_command.MappingCommand",
            parse_function=parse_mapping_command,
            alt_regex=r"(Mapping|Map)([ _]" + regex_cplex_var + "[ _]" + regex_cplex_var + ")?")

]


valid_v2_command_names: List[str] = [label for cmd in commands if cmd.is_v2 for label in cmd.allowed_names]



from openpyxl.worksheet.worksheet import Worksheet

from nexinfosys import IssuesLabelContentTripleType, AreaTupleType
from nexinfosys.command_generators import parser_field_parsers
from nexinfosys.command_generators.parser_spreadsheet_utils import obtain_rectangular_submatrices, \
    binary_mask_from_worksheet
from nexinfosys.common.helper import strcmp


def parse_scale_conversion_command(sh: Worksheet, area: AreaTupleType, name: str = None) -> IssuesLabelContentTripleType:
    """
    Analyze the input area
    Obtain the numerical part
    Read a row above and a column to the left, looking for source (left col) and target (row above) factor types

    FactorTypes do not need to exist previously, they can be created

    :param sh: Input worksheet
    :param area: Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the
    command is present
    :return: list of issues (issue_type, message), command label, command content
    """

    def get_subrow(r, c1, c2):
        lst = []
        # To deal with combined cell ranges, store "previous" value, and if "" is found, assume it is a merged cell
        previous = None
        for c in range(c1, c2):
            v = sh.cell(row=r, column=c).value
            if not v:
                if previous:
                    lst.append(previous)
                else:
                    lst.append("")
            else:
                previous = v
                lst.append(v)

        return lst

    def get_subcolumn(c, r1, r2):
        lst = []
        # To deal with combined cell ranges, store "previous" value, and if "" is found, assume it is a merged cell
        # !!! This may not be correct at all times: when a cell is intentionally left blank
        # To solve this, use "sh.merged_cell_ranges" to check if the current cell (r, c) is inside a range
        previous = None
        for r in range(r1, r2):
            v = sh.cell(row=r, column=c).value
            if not v:
                if previous:
                    lst.append(previous)
                else:
                    lst.append("")
            else:
                previous = v
                lst.append(v)
        return lst

    # ---------------------------------------------

    some_error = False
    issues = []

    # Detect the matrix defining scales
    m = binary_mask_from_worksheet(sh, True)  # "True" is to focus on cells containing numbers
    # Locate the matrix with numbers. Assume this defines the labels to consider, they will be around the matrix
    t = obtain_rectangular_submatrices(m)[0]  # Take just the first tuple: U=t[0], D=t[1], L=t[2], R=t[3]
    t = (t[0]+1, t[1]+1, t[2]+1, t[3]+1)  # The previous calculation is done using Numpy, so it is Zero based. Correct this

    # Obtain the factor type names in the subrow on top of the matrix
    subrow = get_subrow(t[0]-1, t[2], t[3])
    # Obtain the factor type names in the subcolumn to the left of the matrix
    subcol = get_subcolumn(t[2]-1, t[0], t[1])

    # Check that we have valid factor type names
    for ft in subrow+subcol:
        try:
            parser_field_parsers.string_to_ast(parser_field_parsers.simple_h_name, ft)
        except:
            some_error = True
            issues.append((3, "'"+ft+"' is not a valid Factor Type name"))
    if some_error:
        return issues, None, None

    # Scan the matrix, creating scale records
    scales = []
    for i, r in enumerate(range(t[0], t[1])):
        for j, c in enumerate(range(t[2], t[3])):
            v = sh.cell(row=r, column=c).value
            if v:
                if not isinstance(v, str):
                    v = str(v)
                # Origin factor
                origin = subcol[i]
                # Destination factor
                destination = subrow[j]
                if strcmp(origin, destination):
                    issues.append((3, "A change of scale to the same factor type ("+origin+") is not allowed"))
                else:
                    try:
                        parser_field_parsers.string_to_ast(parser_field_parsers.expression_with_parameters, v)
                        # Add the scale
                        scales.append(dict(origin=origin, destination=destination, scale=v))
                    except:
                        issues.append((3, "The expression '"+v+"' at the intersection of factor types " + origin + " and " + destination + " is syntactically incorrect"))

    content = {"origin_factor_types": subcol,
               "destination_factor_types": subrow,
               "scales": scales
               }

    return issues, None, content



from typing import List, Tuple, Dict

from nexinfosys.common.helper import create_dictionary, load_dataset
from nexinfosys.ie_imports.data_source_manager import IDataSourceManager, filter_dataset_into_dataframe
from nexinfosys.models.statistical_datasets import Dataset, DataSource, Database


class AdHocDatasets(IDataSourceManager):
    """
    Datasets in a file (external, or the file currently being analyzed) with format defined by Magic project
    """
    def __init__(self, datasets_list: List[Dataset]):
        self._registry = None
        self.initialize_datasets_registry(datasets_list)

    def initialize_datasets_registry(self, datasets_list: Dict[str, Dataset]):
        """
        Receive a list of the datasets and make a copy

        :param datasets_list:
        :return: None
        """
        self._registry = create_dictionary()
        for ds_name, ds in datasets_list.items():
            self.register_dataset(ds.code, ds)

    def register_dataset(self, name, ds):
        self._registry[name] = ds

    def get_name(self) -> str:
        """ Source name """
        return self.get_datasource().name

    def get_datasource(self) -> DataSource:
        """ Data source """
        src = DataSource()
        src.name = "AdHoc"
        src.description = "A special, ad-hoc, data source, providing datasets elaborated inside an execution. Datasets are local to the execution."
        return src

    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        db = Database()
        db.code = ""
        db.description = "AdHoc is a database itself"
        return [db]

    def get_datasets(self, database=None) -> list:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """

        lst = []
        for d in self._registry:
            lst.append((d, self._registry[d].description))  # [(name, description)]

        return lst

    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """
        return self._registry[dataset]
        # return self.etl_dataset(dataset, update=False)

    def etl_full_database(self, database=None, update=False):
        pass

    def etl_dataset(self, dataset, update=False) -> str:
        """
        Read dataset data and metadata into NIS databases

        :param url:
        :param local_filename:
        :param update:
        :return: String with full file name
        """
        pass

    def get_dataset_filtered(self, dataset: str, dataset_params: List[Tuple]) -> Dataset:
        """ This method has to consider the last dataset download, to re"""
        # Read dataset structure
        # TODO - CLONE "ds"
        ds = self.get_dataset_structure(None, dataset)

        if ds.data is None:
            df = load_dataset(ds.attributes["_location"])
        else:
            df = ds.data

        # Create index with
        idx_cols = []
        for c in ds.dimensions:
            if not c.is_measure:
                idx_cols.append(c.code)

        df = df.set_index(idx_cols)

        # Filter it using generic Pandas filtering capabilities
        if dataset_params.get("StartPeriod") and dataset_params.get("EndPeriod"):
            years = [str(y) for y in range(int(dataset_params["StartPeriod"][0]), int(dataset_params["EndPeriod"][0])+1)]
            dataset_params["year"] = years
            del dataset_params["StartPeriod"]
            del dataset_params["EndPeriod"]
        ds.data = filter_dataset_into_dataframe(df, dataset_params, dataset)

        return ds

    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass





import json
import logging
import re
from typing import Optional, List, Dict, Any, Union, Tuple, Set

from nexinfosys import ExecutableCommandIssuesPairType, Command, CommandField, IssuesOutputPairType, case_sensitive
from nexinfosys.command_definitions import commands
from nexinfosys.command_executors.execution_helpers import classify_variables2
from nexinfosys.command_generators import IType, IssueLocation, Issue, parser_field_parsers
from nexinfosys.command_generators.parser_ast_evaluators import dictionary_from_key_value_list, ast_evaluator
from nexinfosys.command_generators.parser_field_parsers import arith_boolean_expression
from nexinfosys.common.helper import first, PartialRetrievalDictionary, head, strcmp, create_dictionary
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects, State
from nexinfosys.models.musiasem_concepts import Processor, Factor, FactorType, Parameter, Hierarchy, \
    FactorTypesRelationUnidirectionalLinearTransformObservation
from nexinfosys.models.musiasem_concepts_helper import find_processor_by_name, find_factor_types_transform_relation


class CommandExecutionError(Exception):
    pass


def subrow_issue_message(subrow=None):
    return f" (expanded row: {subrow})" if subrow else ""


class BasicCommand(IExecutableCommand):
    def __init__(self, name: str, command_fields: List[CommandField]):
        self._name = name
        self._content: Dict = {}
        self._command_name = ""
        self._command_fields = command_fields

        # Convenience
        self._glb_idx: Optional[PartialRetrievalDictionary] = None
        self._hierarchies = None
        self._datasets = None
        self._mappings = None
        self._parameters: Optional[List[Parameter]] = None
        self._state: Optional[State] = None
        # Execution state per command
        self._issues: List[Issue] = []
        # Execution state per row
        self._current_row_number: Optional[int] = None
        self._fields: Dict[str, Any] = {}

    def _init_execution_state(self, state: Optional["State"] = None) -> None:
        self._issues = []
        self._current_row_number = None
        self._glb_idx = None
        self._hierarchies = None
        self._p_sets = None
        self._datasets = None
        self._mappings = None
        self._parameters = None
        self._fields = {}
        self._state = None

        if state:
            self._glb_idx, self._p_sets, self._hierarchies, self._datasets, self._mappings = get_case_study_registry_objects(state)
            self._parameters = [p.name for p in self._glb_idx.get(Parameter.partial_key())]
            self._state = state

    def _process_row(self, fields: Dict[str, Any], subrow=None) -> None:
        """This is the only method subclasses need to define. See current subclasses as examples"""
        pass

    def _get_command_fields_values(self, row: Dict[str, Any]) -> Dict[str, Any]:
        return {f.name: row.get(f.name, f.default_value) for f in self._command_fields}

    def _check_all_mandatory_fields_have_values(self) -> None:
        empty_fields: List[str] = [f.name
                                   for f in self._command_fields
                                   if f.mandatory and self._fields[f.name] is None]

        if len(empty_fields) > 0:
            raise CommandExecutionError(f"Mandatory field/s '{', '.join(empty_fields)}' is/are empty.")

    def _add_issue(self, itype: IType, description: str) -> None:
        self._issues.append(
            Issue(itype=itype,
                  description=description,
                  location=IssueLocation(sheet_name=self._command_name,
                                         row=self._current_row_number, column=None))
        )
        return

    def _init_and_process_row(self, row: Dict[str, Any]) -> None:
        def obtain_dictionary_with_not_expandable_fields(d):
            output = {}
            for k, v in d.items():
                if v is None or "{" not in v:
                    output[k] = v
            return output

        self._current_row_number = row["_row"]
        self._fields = self._get_command_fields_values(row)
        tmp_fields = self._fields
        self._check_all_mandatory_fields_have_values()
        # If expandable, do it now
        expandable = row["_expandable"]
        if expandable:
            # Extract variables
            state = State()
            issues = []
            asts = {}
            referenced_variables = create_dictionary()
            for e in expandable:
                ast = parser_field_parsers.string_to_ast(arith_boolean_expression, e)
                c_name = f"{{{e}}}"
                asts[c_name] = ast
                res, vars = ast_evaluator(ast, state, None, issues, atomic_h_names=True)
                for v in vars:
                    referenced_variables[v] = None

            res = classify_variables2(referenced_variables.keys(), self._datasets, self._hierarchies, self._parameters)
            ds_list = res["datasets"]
            ds_concepts = res["ds_concepts"]
            h_list = res["hierarchies"]
            if len(ds_list) >= 1 and len(h_list) >= 1:
                self._add_issue(itype=IType.ERROR,
                                description="Dataset(s): " + ", ".join(
                                        [d.name for d in ds_list]) + ", and hierarchy(ies): " + ", ".join([h.name for h in
                                                                                                           h_list]) + ", have been specified. Only a single dataset is supported.")
                return
            elif len(ds_list) > 1:
                self._add_issue(itype=IType.ERROR,
                                description="More than one dataset has been specified: " + ", ".join(
                                        [d.name for d in ds_list]) + ", just one dataset is supported.")
                return
            elif len(h_list) > 0:
                self._add_issue(itype=IType.ERROR,
                                description="One or more hierarchies have been specified: " + ", ".join(
                                        [h.name for h in h_list])
                                )
                return
            if len(ds_list) == 1:  # Expand dataset
                ds = ds_list[0]
                measure_requested = False
                all_dimensions = set([c.code for c in ds.dimensions if not c.is_measure])
                requested_dimensions = set()
                requested_measures = set()
                for con in ds_concepts:
                    found = False
                    for c in ds.dimensions:
                        if strcmp(c.code, con):
                            found = True
                            if c.is_measure:
                                measure_requested = True
                                requested_measures.add(c.code)
                            else:  # Dimension
                                all_dimensions.remove(c.code)
                                requested_dimensions.add(c.code)
                    if not found:
                        self._add_issue(
                            itype=IType.ERROR,
                            description=f"The concept '{{{ds.code}.{con}}}' is not in the dataset '{ds.code}'"
                        )
                        return
                ds_concepts = list(requested_measures)
                ds_concepts.extend(list(requested_dimensions))
                all_dimensions_requested = len(all_dimensions) == 0

                if measure_requested and not all_dimensions_requested:
                    self._add_issue(IType.ERROR,
                                    f"It is not possible to use a measure ({', '.join(requested_measures)}), if not all dimensions are used "
                                    f"(cannot assume implicit aggregation). Dimensions not used: {', '.join(all_dimensions)}")
                    return
                elif not measure_requested and not all_dimensions_requested:
                    # Reduce the Dataframe to unique tuples of the specified dimensions
                    # TODO Consider the current case -sensitive or not-sensitive-
                    data = ds.data[list(requested_dimensions)].drop_duplicates()
                else:  # Take the dataset as-is
                    data = ds.data

                # Remove Index, and do it NOT-INPLACE
                data = data.reset_index()

                # Drop rows with empty dimension value
                # import numpy as np
                # data = data.replace(r'^\s*$', np.NaN, regex=True)
                # data.dropna(subset=requested_dimensions, inplace=True)

                const_dict = obtain_dictionary_with_not_expandable_fields(self._fields)  # row?
                var_dict = set([f for f in self._fields.keys() if f not in const_dict])

                re_concepts = {}
                for c in ds_concepts:
                    c_name = f"{{{ds.code}.{c}}}"
                    if case_sensitive:
                        re_concepts[c_name] = re.compile(c_name)
                    else:
                        re_concepts[c_name] = re.compile(c_name, re.IGNORECASE)

                location = IssueLocation(sheet_name=self._command_name, row=self._current_row_number, column=None)
                already_parsed_fields = set(const_dict.keys())
                for ds_row, row2 in enumerate(data.iterrows()):  # Each row in the dataset
                    # Initialize constant values (those with no "{..}" expressions)
                    row3 = const_dict.copy()
                    # Prepare state to evaluate functions
                    state = State()
                    for c in ds_concepts:
                        state.set(f"{ds.code}.{c}", str(row2[1][c]))
                    state.set("_glb_idx", self._glb_idx)  # Pass PartialRetrievalDictionary to the evaluator. For functions needing it

                    # Evaluate all functions
                    expressions = {}
                    for e, ast in asts.items():
                        res, vars = ast_evaluator(ast, state, None, issues, atomic_h_names=True)
                        expressions[e] = res
                    # Expansion into var_dict
                    for f in var_dict:
                        v = self._fields[f]  # Initial value
                        for item in sorted(expressions.keys(), key=len, reverse=True):
                            v = v.replace(item, expressions[item])
                        row3[f] = v

                    # # Concepts change dictionary
                    # concepts = {}
                    # for c in ds_concepts:
                    #     concepts[f"{{{ds.code}.{c}}}"] = str(row2[1][c])
                    # # Expansion into var_dict
                    # for f in var_dict:
                    #     v = self._fields[f]  # Initial value
                    #     for item in sorted(concepts.keys(), key=len, reverse=True):
                    #         v = re_concepts[item].sub(concepts[item], v)
                    #     row3[f] = v

                    # Syntactic verification of the resulting expansion
                    processable, tmp_issues = parse_cmd_row_dict(self._serialization_type,
                                                                 row3,
                                                                 already_parsed_fields,
                                                                 location)
                    if len(tmp_issues) > 0:
                        self._issues.extend(tmp_issues)
                    # Process row
                    if processable:
                        self._fields = row3
                        self._process_row(row3, ds_row)
                        self._fields = tmp_fields
            elif len(h_list) == 1:  # Expand hierarchy
                pass
        else:
            self._process_row(self._fields)  # Process row

    def execute(self, state: "State") -> IssuesOutputPairType:
        """Main entry point"""
        self._init_execution_state(state)

        for row in self._content["items"]:
            try:
                self._init_and_process_row(row)
            except CommandExecutionError as e:
                self._add_issue(IType.ERROR, str(e))

        return self._issues, None

    def estimate_execution_time(self) -> int:
        return 0

    def json_serialize(self) -> Dict:
        """Directly return the metadata dictionary"""
        return self._content

    def json_deserialize(self, json_input: Union[dict, str, bytes, bytearray]) -> List[Issue]:
        # TODO Check validity
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        self._command_name = self._content["command_name"]

        return []

    def _get_processor_from_field(self, field_name: str) -> Processor:
        processor_name = self._fields[field_name]
        try:
            processor = find_processor_by_name(state=self._glb_idx, processor_name=processor_name)
        except Exception as e:
            raise CommandExecutionError(e)

        if not processor:
            raise CommandExecutionError(f"The processor '{processor_name}' defined in field '{field_name}' "
                                        f"has not been previously declared.")
        return processor

    def _get_interface_from_field(self, field_name: str, processor: Processor) -> Factor:
        interface_name = self._fields[field_name]

        if interface_name is None:
            raise CommandExecutionError(f"No interface has been defined for field '{field_name}'.")

        interface = processor.factors_find(interface_name)
        if not interface:
            raise CommandExecutionError(f"The interface '{interface_name}' has not been found in "
                                        f"processor '{processor.name}'.")

        return interface

    def _get_factor_type_from_field(self, hierarchy_field_name: str, interface_type_field_name: str) -> FactorType:
        interface_type_name = self._fields[interface_type_field_name]
        if not interface_type_name:
            raise CommandExecutionError(f"The field '{interface_type_field_name}' has not been specified")

        # Check if FactorType exists
        interface_types = self._glb_idx.get(FactorType.partial_key(interface_type_name))

        if len(interface_types) == 1:
            return interface_types[0]
        elif len(interface_types) == 0:
            raise CommandExecutionError(f"The interface type '{interface_type_name}' has not been found")
        else:
            hierarchy_name = self._fields[hierarchy_field_name]
            if not hierarchy_name:
                raise CommandExecutionError(f"The field '{hierarchy_field_name}' has not been specified and "
                                            f"the interface type '{interface_type_name}' is not unique")

            interface_type = first(interface_types, lambda t: strcmp(t.hierarchy.name, hierarchy_name))
            if not interface_type:
                raise CommandExecutionError(f"The interface type '{interface_type_name}' has not been found in "
                                            f"hierarchy '{hierarchy_name}'")

            return interface_type

    def _get_factor_types_from_field(self, hierarchy_field_name: str, interface_type_field_name: str) -> List[FactorType]:
        """ Possibly obtain not only one but many InterfaceTypes """

        hierarchy_name = self._fields[hierarchy_field_name]
        interface_type_name = self._fields[interface_type_field_name]

        if not interface_type_name and not hierarchy_name:
            raise CommandExecutionError(f"No hierarchy nor interface type have been specified. At least specify one of them.")
        elif interface_type_name and hierarchy_name:
            interface_types = self._glb_idx.get(FactorType.partial_key(interface_type_name))
            if len(interface_types) == 1:
                return [interface_types[0]]
            elif len(interface_types) == 0:
                raise CommandExecutionError(f"The interface type '{interface_type_name}' has not been found")
            else:
                hierarchy_name = self._fields[hierarchy_field_name]
                if not hierarchy_name:
                    raise CommandExecutionError(f"The field '{hierarchy_field_name}' has not been specified and "
                                                f"the interface type '{interface_type_name}' is not unique")

                interface_type = first(interface_types, lambda t: strcmp(t.hierarchy.name, hierarchy_name))
                if not interface_type:
                    raise CommandExecutionError(f"The interface type '{interface_type_name}' has not been found in "
                                                f"hierarchy '{hierarchy_name}'")

                return [interface_type]
        elif interface_type_name and not hierarchy_name:
            interface_types = self._glb_idx.get(FactorType.partial_key(interface_type_name))
            if len(interface_types) == 1:
                return [interface_types[0]]
            elif len(interface_types) == 0:
                raise CommandExecutionError(f"The interface type '{interface_type_name}' has not been found")
            else:
                raise CommandExecutionError(f"The field '{hierarchy_field_name}' has not been specified and "
                                            f"the interface type '{interface_type_name}' is not unique")
        elif not interface_type_name and hierarchy_name:
            hie = self._glb_idx.get(Hierarchy.partial_key(hierarchy_name))
            if len(hie) == 1:
                # All children of "hierarchy_name"
                return [v for v in hie[0].codes.values()]
            elif len(hie) == 0:
                raise CommandExecutionError(f"The InterfaceTypes hierarchy '{hierarchy_name}' has not been found")
            else:
                raise CommandExecutionError(f"The InterfaceTypes hierarchy '{hierarchy_name}' has been found multiple times!!")

        # Check if FactorType exists
        interface_types = self._glb_idx.get(FactorType.partial_key(interface_type_name))

        if len(interface_types) == 1:
            return interface_types[0]
        elif len(interface_types) == 0:
            raise CommandExecutionError(f"The interface type '{interface_type_name}' has not been found")
        else:
            hierarchy_name = self._fields[hierarchy_field_name]
            if not hierarchy_name:
                raise CommandExecutionError(f"The field '{hierarchy_field_name}' has not been specified and "
                                            f"the interface type '{interface_type_name}' is not unique")

            interface_type = first(interface_types, lambda t: strcmp(t.hierarchy.name, hierarchy_name))
            if not interface_type:
                raise CommandExecutionError(f"The interface type '{interface_type_name}' has not been found in "
                                            f"hierarchy '{hierarchy_name}'")

            return interface_type

    def _get_attributes_from_field(self, field_name: str) -> Dict:
        attributes_list = self._fields[field_name]
        attributes = {}
        if attributes_list:
            try:
                attributes = dictionary_from_key_value_list(attributes_list, self._glb_idx)
            except Exception as e:
                raise CommandExecutionError(str(e))

        return attributes

    def _get_interface_types_transform(self,
                                       source_interface_type: FactorType, source_processor: Processor,
                                       target_interface_type: FactorType, target_processor: Processor,
                                       subrow=None) \
            -> FactorTypesRelationUnidirectionalLinearTransformObservation:
        """Check if a transformation between interfaces has been specified"""

        interface_types_transforms = find_factor_types_transform_relation(
            self._glb_idx, source_interface_type, target_interface_type, source_processor, target_processor)

        if len(interface_types_transforms) == 0:
            raise CommandExecutionError(f"Interface types are not the same (and transformation from one "
                                        f"to the other cannot be performed). Origin: "
                                        f"{source_interface_type.name}; Target: {target_interface_type.name}"+subrow_issue_message(subrow))
        elif len(interface_types_transforms) > 1:
            raise CommandExecutionError(
                f"Multiple transformations can be applied between interfaces. Origin: "
                f"{source_interface_type.name}; Target: {target_interface_type.name}"+subrow_issue_message(subrow))

        return interface_types_transforms[0]


class ParseException(Exception):
    pass


def parse_cmd_row_dict(cmd_name: str, row: Dict[str, str], already_parsed_fields: Set[str], location: IssueLocation) -> Tuple[bool, List[Issue]]:
    """
    Parse a row (as a dictionary) from a command
    It is used after expansion of "macros"

    :param cmd_name: Name of command
    :param row: A dictionary containing the values to parse syntactically. Keys are field names, Values are field values
    :param already_parsed_fields: Set of fields already known to be syntactically valid
    :param location: IssueLocation object to use when creating Issues
    :return: A tuple: a boolean (True if the row can be used, otherwise False) and a list of Issues
    """

    issues: List[Issue] = []

    from nexinfosys.command_field_definitions import command_fields
    field_defs_dict = {f.name: f for f in command_fields[cmd_name]}
    mandatory_not_found = set([c.name for c in command_fields[cmd_name] if c.mandatory and isinstance(c.mandatory, bool)])
    # logging.debug(mandatory_not_found)
    complex_mandatory_cols = [c for c in command_fields[cmd_name] if isinstance(c.mandatory, str)]
    may_append = True
    complex_row = False
    for field_name, field_value in row.items():
        field_def = field_defs_dict.get(field_name)
        if not field_def:
            return ParseException(f"Field {field_name} not found for command {cmd_name}")

        if field_value is not None:
            if not isinstance(field_value, str):
                field_value = str(field_value)
            field_value = field_value.strip()
        else:
            continue

        # Parse the field
        if field_def.allowed_values:
            if field_value.lower() not in [v.lower() for v in field_def.allowed_values]:  # TODO Case insensitive CI
                issues.append(
                    Issue(itype=IType.ERROR,
                          description=f"Field '{field_name}' of command '{cmd_name}' has invalid value '{field_value}'."
                                      f" Allowed values are: {', '.join(field_def.allowed_values)}.",
                          location=location))
                may_append = False
            else:
                pass  # OK
        else:  # Instead of a list of values, check if a syntactic rule is met by the value
            if field_def.parser:  # Parse, just check syntax (do not store the AST)
                try:
                    if field_name not in already_parsed_fields:
                        ast = parser_field_parsers.string_to_ast(field_def.parser, field_value)
                        # Rules are in charge of informing if the result is expandable and if it complex
                        if "expandable" in ast and ast["expandable"]:
                            issues.append(
                                Issue(itype=IType.ERROR,
                                      description=f"Field '{field_name}' of command '{cmd_name}' cannot be expandable again.",
                                      location=location)
                            )
                            may_append = False
                        if "complex" in ast and ast["complex"]:
                            complex_row = True
                except:
                    issues.append(Issue(itype=IType.ERROR,
                                        description=f"The value in field '{field_name}' of command '{cmd_name}' "
                                        f"is not syntactically correct. Entered: {field_value}",
                                        location=location))
                    may_append = False
            else:
                pass  # Valid

        if field_def.name in mandatory_not_found:
            mandatory_not_found.discard(field_def.name)

    # MODIFY INPUT Dictionary with this new Key
    if complex_row:
        row["_complex"] = complex_row

    # Append if all mandatory fields have been filled
    if len(mandatory_not_found) > 0:
        issues.append(Issue(itype=IType.ERROR,
                            description=f"Mandatory columns: {', '.join(mandatory_not_found)} have not been specified",
                            location=location))
        may_append = False

    # Check varying mandatory fields (fields depending on the value of other fields)
    for c in complex_mandatory_cols:
        field_def = c.name  # next(c2 for c2 in col_map if strcmp(c.name, c2.name))
        if isinstance(c.mandatory, str):
            # Evaluate
            mandatory = eval(c.mandatory, None, row)
            may_append = (mandatory and field_def in row) or (not mandatory)
            if mandatory and field_def not in row:
                issues.append(Issue(itype=IType.ERROR,
                                    description="Mandatory column: " + field_def + " has not been specified",
                                    location=location))
                may_append = False

    return may_append, issues


def create_command(cmd_type, name, json_input, source_block=None) -> ExecutableCommandIssuesPairType:
    """
    Factory creating and initializing a command from its type, optional name and parameters

    :param cmd_type: String describing the type of command, as found in the interchange JSON format
    :param name: An optional name (label) for the command
    :param json_input: Parameters specific of the command type (each command knows how to interpret,
                       validate and integrate them)
    :param source_block: String defining the name of the origin block (in a spreadsheet is the worksheet name, but other block types could appear)
    :return: The instance of the command and the issues creating it
    :raise
    """
    cmd: Optional[Command] = first(commands, condition=lambda c: c.name == cmd_type)

    if cmd:
        if cmd.execution_class:
            exec_cmd: "IExecutableCommand" = cmd.execution_class(name)  # Reflective call
            exec_cmd._serialization_type = cmd_type  # Injected attribute. Used later for serialization
            exec_cmd._serialization_label = name  # Injected attribute. Used later for serialization
            exec_cmd._source_block_name = source_block
            if isinstance(json_input, (str, dict, list)):
                if json_input != {}:
                    issues = exec_cmd.json_deserialize(json_input)
                else:
                    issues = []
            else:
                # NO SPECIFICATION
                raise Exception("The command '" + cmd_type + " " + name if name else "<unnamed>" + " does not have a specification.")
            return exec_cmd, issues  # Return the command and the issues found during the deserialization
        else:
            return None, []  # No execution class. Currently "import_commands" and "list_of_commands"
    else:
        # UNKNOWN COMMAND
        raise Exception("Unknown command type: " + cmd_type)



import io
import logging
import traceback
from typing import List, Tuple, Dict

import pandas as pd
import xlrd

import nexinfosys
from nexinfosys.command_generators import IType
from nexinfosys.common.helper import any_error_issue, download_file
from nexinfosys.model_services import get_case_study_registry_objects, State
from nexinfosys.model_services.workspace import InteractiveSession, prepare_and_solve_model
from nexinfosys.models.musiasem_methodology_support import DBSession
from nexinfosys.initialization import initialize_databases, get_parameters_in_state, get_scenarios_in_state, \
    register_external_datasources, get_graph_from_state, get_dataset_from_state, get_model, get_geolayer, get_ontology, \
    validate_command, command_field_help, comm_help
from nexinfosys import initialize_configuration
from nexinfosys.serialization import serialize_state, deserialize_state


class NIS:
    def __init__(self):
        logging.basicConfig(level=logging.DEBUG)
        self._isession = None
        self._rsession = None
        self._dataframe_names = []  # type: List[str]
        self._dataframes = []  # type: List[pd.DataFrame]
        self._issues = None
        self._state = None
        initialize_configuration()
        # Disable optimizations
        nexinfosys.set_global_configuration_variable("ENABLE_CYTHON_OPTIMIZATIONS", False)
        initialize_databases()
        nexinfosys.data_source_manager = register_external_datasources()

    @property
    def dataframes(self):
        return self._dataframes

    @property
    def dataframe_names(self):
        return self._dataframe_names

    # --------------- SESSION ---------------

    def open_session(self, reset_commands=False, state: State=None):
        self._isession = InteractiveSession(DBSession)
        if state:
            self._isession.state = state
        self._isession.identify({"user": "test_user", "password": None}, testing=True)
        self._rsession = self._isession.open_reproducible_session(case_study_version_uuid=None,
                                                                  recover_previous_state=False,
                                                                  cr_new=True,
                                                                  allow_saving=False)

        if reset_commands:
            # Clear dataframes
            self.reset_commands()

        return self._isession

    def close_session(self):
        # Close reproducible session
        uuid_, v_uuid, cs_uuid = self._isession.close_reproducible_session(issues=None, output=None, save=False, from_web_service=False, cs_uuid=None, cs_name=None)
        self._isession.quit()
        self._isession = None
        self._rsession = None

    # --------------- SUBMISSION PREPARATION ---------------

    def reset_commands(self):
        self._dataframe_names.clear()
        self._dataframes.clear()

    def load_workbook(self, fname, wv_user=None, wv_password=None, wv_host_name=None):
        """

        :param fname:
        :param wv_user: In case of use of WebDav server, user name
        :param wv_password: In case of use of WebDav server, password
        :param wv_host_name: In case of use of WebDav server, host name
        :return: Number of added DataFrames
        """
        # Load a XLSX workbook into memory, as dataframes
        bytes_io = download_file(fname, wv_user, wv_password, wv_host_name)
        xl = pd.ExcelFile(bytes_io, engine='openpyxl')
        # xl = pd.ExcelFile(xlrd.open_workbook(file_contents=bytes_io.getvalue()), engine="xlrd")
        cont = 0
        for sheet_name in xl.sheet_names:
            df = xl.parse(sheet_name, header=0)
            # Manage columns
            cols = []
            for col in df.columns:
                col_parts = col.split(".")
                if col.lower().startswith("unnamed"):
                    cols.append("")
                elif len(col_parts) > 1:
                    try:
                        int(col_parts[1])  # This is the case of "col.1"
                        cols.append(col_parts[0])
                    except:  # This is the case of "col_part.col_part" (second part is string)
                        cols.append(col)
                else:
                    cols.append(col)

            df.columns = cols
            self._dataframes.append(df)
            self._dataframe_names.append(sheet_name)
            cont += 1

        return cont

    def _generate_in_memory_excel(self):
        output = io.BytesIO()
        writer = pd.ExcelWriter(output, engine="xlsxwriter")
        for idx, sheet_name in enumerate(self._dataframe_names):
            header = sheet_name.lower() != "metadata"
            self._dataframes[idx].to_excel(writer, sheet_name=sheet_name, index=False)
        writer.save()
        return output.getvalue()

    def append_command(self, name: str, df: pd.DataFrame):
        for n in self._dataframe_names:
            if n.lower() == name.lower():
                raise Exception("Duplicated name '"+name+"' for a command.")

        if not isinstance(df, pd.DataFrame):
            raise Exception("'df' parameter must be a Pandas DataFrame")

        self._dataframes.append(df)
        self._dataframe_names.append(name)

    def get_external_datasets(self) -> Dict[str, Tuple[str, str]]:
        """
        Obtain a list of datasets usable in DatasetQry command

        :return: Dict(dataset_code, (source, dataset_description))
        """
        pass

    def get_external_dataset_structure(self, dataset_code) -> Tuple[Dict[str, List[str]], List[str]]:
        """
        Obtain the dimensions

        :param dataset_code:
        :return: Tuple with: Dimensions in dictionary (dim_name, list of codes), List of measure names
        """
        pass

    def get_datasetqry_dataframe(self, dataset_code, filter, out_dimensions, out_measures: List[Tuple[str, str, str]], out_dataset_name) -> pd.DataFrame:
        """
        Obtain a pd.DataFrame with a DatasetQry command

        :param dataset_code: code of the external dataset to use as input
        :param filter: dictionary with dimension name and list of codes passing the filter
        :param out_dimensions: List of dimensions in the output
        :param out_measures: List of measures in the output dataset. Each measure: aggregation function, measure to aggregate, result measure name
        :param out_dataset_name: Name of the output dataset
        :return: pd.DataFrame compatible with NIS format (append_command method can be invoked immediately, then submit method and the "after submission" methods to obtain the resulting dataset)
        """
        pass

    # --------------- SUBMISSION ---------------

    def submit(self, ignore_imports=False) -> List:
        if len(self._dataframes) == 0:
            raise Exception("The list of dataframes is empty. Cannot submit.")

        # A reproducible session must be open
        if self._isession:
            # Add system-level entities from JSON definition in "default_cmds"
            self._isession.register_andor_execute_command_generator("json", "application/json", nexinfosys.default_cmds, False, True)

            # PARSE AND BUILD!!!
            generator_type = "spreadsheet"
            content_type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
            buffer = self._generate_in_memory_excel()
            logging.debug(len(buffer))
            execute = True
            register = False

            try:
                ret = self._isession.register_andor_execute_command_generator(generator_type, content_type, buffer, register, execute, ignore_imports)
                if isinstance(ret, tuple):
                    issues = ret[0]
                else:
                    issues = []

                # TODO CHECK SEMANTIC INCONSISTENCIES. Referred to values in Interfaces use either Parameters

            except Exception as e:
                traceback.print_exc()  # Print the Exception to std output
                # Obtain trace as string; split lines in string; take the last three lines
                tmp = traceback.format_exc().splitlines()
                for i in range(len(tmp) - 3, 0, -2):
                    if tmp[i].find("nexinfosys") != -1:
                        tmp = [tmp[-1], tmp[i], tmp[i + 1]]
                        break
                else:
                    tmp = [tmp[-1], "Nexinfosys module not found", "Line not found"]
                exc_info = ' :: '.join([s.strip() for s in tmp])
                # Error Issue with the extracted Exception text
                issues = [nexinfosys.Issue(itype=IType.ERROR,
                                           description=f"UNCONTROLLED CONDITION: {exc_info}. Please, contact the development team.",
                                           location=None)]

            # STORE the issues
            self._issues = issues
            self._state = self._isession.state
            return self._issues
        else:
            raise Exception("Call 'open_session' before submitting")

    def solve(self) -> List:
        # A reproducible session must be open
        if self._isession:
            try:
                # SOLVE !!!!
                if not any_error_issue(self._issues):
                    issues2 = prepare_and_solve_model(self._isession.state)
                    self._issues.extend(issues2)
            except Exception as e:
                traceback.print_exc()  # Print the Exception to std output
                # Obtain trace as string; split lines in string; take the last three lines
                tmp = traceback.format_exc().splitlines()
                for i in range(len(tmp) - 3, 0, -2):
                    if tmp[i].find("nexinfosys") != -1:
                        tmp = [tmp[-1], tmp[i], tmp[i + 1]]
                        break
                else:
                    tmp = [tmp[-1], "Nexinfosys module not found", "Line not found"]
                exc_info = ' :: '.join([s.strip() for s in tmp])
                # Error Issue with the extracted Exception text
                issues = [nexinfosys.Issue(itype=IType.ERROR,
                                           description=f"UNCONTROLLED CONDITION: {exc_info}. Please, contact the development team.",
                                           location=None)]

            self._state = self._isession.state

            return self._issues
        else:
            raise Exception("Call 'open_session' before submitting")

    def submit_and_solve(self):
        self.submit()
        return self.solve()

    def get_state(self):
        return self._state

    # --------------- AFTER SUBMISSION ---------------

    def query_parameters(self):
        """
        Return the parameters.
        For each parameter return: name, type (Number, Code, Boolean, String) and the domain (range for Number, list of codes for Code)

        :return:
        """
        if self._isession:
            r = get_parameters_in_state(self._isession.state)
            return r
        else:
            raise Exception("Call 'open_session' before querying for available parameters")

    def query_scenarios(self):
        """
        Return the list of scenarios and their parameters

        :return:
        """
        if self._isession:
            r = get_scenarios_in_state(self._isession.state)
            return r
        else:
            raise Exception("Call 'open_session' before querying for available scenarios")

    def recalculate(self, parameters: dict):
        """
        Once a case study has been submitted, recalculate (RESUBMIT?) with new parameter values

        :param parameters:
        :return:
        """
        if self._isession:
            # Pass the dictionary of parameters
            issues2 = prepare_and_solve_model(self._isession.state, parameters)
            return issues2
        else:
            raise Exception("Call 'open_session' before resubmitting a new scenario")

    def query_available_datasets(self):
        def get_results_in_session(isess: InteractiveSession):
            dataset_formats = ["CSV", "XLSX", "SDMX.json"]  # , "XLSXwithPivotTable", "NISembedded", "NISdetached"]
            graph_formats = ["VisJS", "html", "GML"]  # , "GraphML"]
            ontology_formats = ["OWL"]
            geo_formats = ["GeoJSON"]
            # A reproducible session must be open, signal about it if not
            if isess.reproducible_session_opened():
                if isess.state:
                    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(isess.state)
                    r = [dict(name=k,
                              type="dataset",
                              description=F"{datasets[k].description} [{datasets[k].data.shape[0]} rows, {datasets[k].data.size} cells, "
                                          F"{datasets[k].data.memory_usage(True).sum()} bytes]",
                              formats=[dict(format=f,
                                            name=f"{k}.{f.lower()}")
                                       for f in dataset_formats],

                              ) for k in datasets
                         ] + \
                        [
                            dict(name="interfaces_graph",
                                 type="graph",
                                 description="Graph of Interfaces, Quantities; Scales and Exchanges",
                                 formats=[dict(format=f,
                                               name=F"interfaces_graph.{f.lower()}")
                                          for f in graph_formats]),
                        ] + \
                        [dict(name="processors_graph",
                              type="graph",
                              description="Processors and exchanges graph",
                              formats=[dict(format=f,
                                            url=F"processors_graph.{f.lower()}")
                                       for f in graph_formats])
                         ] + \
                        [dict(name="processors_geolayer",
                              type="geolayer",
                              description="Processors located in a geographic layer",
                              formats=[
                                  dict(format=f, url=F"geolayer.{f.lower()}")
                                  for f in geo_formats]),
                         ] + \
                        [dict(name="Model",
                              type="model",
                              description="Model",
                              formats=[
                                  dict(format=f, url=F"model.{f.lower()}")
                                  for f in ["JSON", "XLSX", "XML"]]),
                         ]
                        # [dict(name="Ontology",
                        #       type="ontology",
                        #       description="OWL ontology",
                        #       formats=[
                        #           dict(format=f, url=F"ontology.{f.lower()}")
                        #           for f in ontology_formats]),
                        #  ]

            return r

        if self._isession:
            # Obtain a list of datasets and the type
            r = get_results_in_session(self._isession)
            if r is not None:
                return r
            else:
                raise Exception("Could not retrieve the available datasets")
        else:
            raise Exception("Call 'open_session' before querying for available datasets")

    def get_results(self, datasets: List[Tuple[str, str, str]]):
        """
        Get one or more results, in the specified format

        List of Tuple: ("type of structure", "dataset name", "output format")

        Example:
          nis.get_results(["dataset", "ds1"])
          nis.get_results([None, "ds1"])  # Guesses the dataset type

        :param datasets:
        :return:
        """
        if self._isession:
            struc_types = {ds["name"]: ds["type"] for ds in self.query_available_datasets()}

            res = []

            for t in datasets:
                struc_type = t[0]
                ds_name = t[1]
                if len(t) == 3:
                    out_format = t[2]
                else:
                    out_format = "bytearray"

                if "." in ds_name:
                    pos = ds_name.find(".")
                    extension = ds_name[pos + 1:]
                    ds_name = ds_name[:pos]
                else:
                    extension = None

                if struc_type is None:
                    struc_type = struc_types.get(ds_name, "")

                if struc_type == "dataset":
                    ds, ctype, ok = get_dataset_from_state(self._isession.state, ds_name, extension, labels_enabled=True)
                elif struc_type == "graph":
                    ds, ctype, ok = get_graph_from_state(self._isession.state, ds_name+"."+extension)
                elif struc_type == "geolayer":
                    ds, ctype, ok = get_geolayer(self._isession.state, extension)
                elif struc_type == "model":
                    ds, ctype, ok = get_model(self._isession.state, extension)
                elif struc_type == "ontology":
                    ds, ctype, ok = get_ontology(self._isession.state, extension)
                else:
                    ds, ctype, ok = None, None, False

                res.append((ds, ctype, ok))
            return res
        else:
            raise Exception("Call 'open_session' before obtaining available datasets")

    # --------------- SYNTAX FUNCTIONS ---------------

    @staticmethod
    def validate_cell(command_type: str, column_name: str, value: str):
        """
        Validate syntax of cell, given the command type and column under which the value is specified

        :param command_type: Command type
        :param column_name: Column in command type
        :param value: Specific cell value
        :return:
        """
        return validate_command(dict(command=command_type, fields={column_name: value}))

    def list_of_available_command_types(self):
        """
        List of commands
        :return:
        """

    def fields_for_command_type(self, command_type: str):
        """
        Fields expected by command type
        :param command_type: Name of t
        :return:
        """

    def examples_for_command_type(self, command_type: str):
        """
        Examples of command type. Returns a list of CSVs or URLs (pointing to case studies using this command type)

        :param command_type: Command type
        :return:
        """

    def description_for_command_type(self, command_type: str):
        """
        A text describing the purpose and semantics of the command type

        :param command_type: Command type
        :return:
        """

    @staticmethod
    def command_fields_help(command_type: str, column_name: str):
        return command_field_help(dict(command=command_type, fields=[column_name]))

    @staticmethod
    def command_help(command_type: str):
        return comm_help(dict(command=command_type))


if __name__ == '__main__':
    nis = NIS()
    nis.open_session(True)
    nis.reset_commands()
    # Load file and execute it
    nis.load_workbook("https://docs.google.com/spreadsheets/d/15NNoP8VjC2jlhktT0A8Y0ljqOoTzgar8l42E5-IRD90/edit#gid=2056162596")
    issues = nis.submit()

    tmp = nis.query_available_datasets()
    print("-------------------")
    print(tmp)
    d = nis.get_results([(tmp[0]["type"], tmp[0]["name"])])
    print(d)

    serial_state = serialize_state(nis.get_state())
    nis.close_session()
    # -------------------
    for i in range(10):
        nis = NIS()
        nis.open_session(True, deserialize_state(serial_state))
        nis.load_workbook("https://docs.google.com/spreadsheets/d/1z8QKYkvlUPLAwo6NYM8rRLMj7_QjPRYOgAsTRI2HwWE/edit#gid=1467907392")
        nis.submit(ignore_imports=True)

        tmp = nis.query_available_datasets()
        print("-------------------")
        print(tmp)
        d = nis.get_results([(tmp[0]["type"], tmp[0]["name"])])
        print(d)

        nis.close_session()

    examples = [
        ("https://docs.google.com/spreadsheets/d/1z8QKYkvlUPLAwo6NYM8rRLMj7_QjPRYOgAsTRI2HwWE/edit#gid=1467907392", False),
        ("https://docs.google.com/spreadsheets/d/12AlJ0tdu2b-cfalNzLqFYfiC-hdDlIv1M1pTE3AfSWY/edit#gid=84311637", False),
        ("/home/rnebot/GoogleDrive/AA_PROPUESTAS/Sentinel/Enviro/examples/01_eurostat_datasets.xlsx", False),
        ("/home/rnebot/GoogleDrive/AA_PROPUESTAS/Sentinel/Enviro/examples/howTo Examples SC-Observers.xlsx", True),
        # ("/home/rnebot/Dropbox/nis-backend/backend_tests/z_input_files/v2/Netherlandsv1ToNISHierarchy2.xlsx", True),
        # ("/home/rnebot/Dropbox/nis-backend/backend_tests/z_input_files/v2/Biofuel_NIS.xlsx", False),
                ]
    nis = NIS()
    for example in examples:
        nis.open_session()
        nis.reset_commands()
        nis.load_workbook(example[0])
        issues = nis.submit()
        if example[1]:
            issues2 = nis.solve()
        tmp = nis.query_available_datasets()
        print(tmp)
        d = nis.get_results([(tmp[0]["type"], tmp[0]["name"])])
        print(d)




import functools

from flask import after_this_request, request


def gzipped(f):
    """
    Decorator to ZIP the response
    """
    @functools.wraps(f)
    def view_func(*args, **kwargs):
        @after_this_request
        def zipper(response):
            accept_encoding = request.headers.get('Accept-Encoding', '')

            if 'gzip' not in accept_encoding.lower():
                return response

            response.direct_passthrough = False

            if (response.status_code < 200 or
                response.status_code >= 300 or
                'Content-Encoding' in response.headers):
                return response
            gzip_buffer = IO()
            gzip_file = gzip.GzipFile(mode='wb',
                                      fileobj=gzip_buffer)
            gzip_file.write(response.data)
            gzip_file.close()

            response.data = gzip_buffer.getvalue()
            response.headers['Content-Encoding'] = 'gzip'
            response.headers['Vary'] = 'Accept-Encoding'
            response.headers['Content-Length'] = len(response.data)

            return response

        return f(*args, **kwargs)

    return view_func



"""
List of parameters. Name, initial value (optional), domain (range), description

"""
from nexinfosys.command_generators import parser_field_parsers
from nexinfosys.common.helper import strcmp, create_dictionary


def parse_parameters_command(sh, area):
    """
    Analyze the input to produce a JSON object with a list of parameters

    :param sh: Input worksheet
    :param area: Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the
    command is present
    :return: list of issues (issue_type, message), command label, command content
    """
    some_error = False
    issues = []

    # Scan the sheet, the first column must be one of the keys of "k_list", following
    # columns can contain repeating values
    col_names = {("name",): "name",
                 ("value",): "value",
                 ("type",): "type",  # category, integer or number
                 ("range", "domain", "interval"): "range",
                 ("group",): "group",  # To group (or filter) parameters
                 ("description", "label", "desc"): "description"
                 }
    # Check columns
    col_map = {}
    for c in range(area[2], area[3]):
        col_name = sh.cell(row=area[0], column=c).value
        for k in col_names:
            if col_name.lower() in k:
                col_map[col_names[k]] = c
                break

    # Map key to a list of values
    content = []  # Dictionary of lists, one per metadata key
    for r in range(area[0]+1, area[1]):
        param = {}
        for k in col_names.values():
            if k not in col_map:
                continue

            value = sh.cell(row=r, column=col_map[k]).value

            if not value:
                continue

            if k == "name":  # Mandatory
                # Check syntax
                try:
                    parser_field_parsers.string_to_ast(parser_field_parsers.simple_ident, value)
                    param[k] = value
                except:
                    some_error = True
                    issues.append((3, "The name specified for the parameter, '" + value + "', is not valid, in row " + str(r) + ". It must be a simple identifier."))
            elif k == "value":  # Optional
                # Check syntax
                try:
                    # TODO Define the parser
                    # basic_elements_parser.string_to_ast(basic_elements_parser.simple_ident, value)
                    param[k] = value
                except:
                    some_error = True
                    issues.append((3, "The Value specified for the parameter, '" + value + "', is not valid, in row " + str(r) + ". It must be an expression."))
            elif k == "type":  # Mandatory
                # Check syntax
                if value.lower().strip() in ("category", "integer", "number", "float"):
                    if value.lower().strip() in ("number", "float"):
                        value = "number"  # "float" --> "number"
                    param[k] = value
                else:
                    some_error = True
                    issues.append((3, "The Type specified for the parameter, '" + value + "', is not valid, in row " + str(r) + ". It must be one of 'category', 'integer', 'number'."))
            elif k == "range":  # Optional
                try:
                    # TODO Define the parser. An interval or a list of possibilities
                    # basic_elements_parser.string_to_ast(basic_elements_parser.simple_ident, value)
                    param[k] = value
                except:
                    some_error = True
                    issues.append((3, "The Range specified for the parameter, '" + value + "', is not valid, in row " + str(r) + ". It must be an interval or a list of categories, depending on the type."))
            elif k == "group":  # Optional
                # Check syntax. A way to group parameters
                try:
                    parser_field_parsers.string_to_ast(parser_field_parsers.simple_ident, value)
                    param[k] = value
                except:
                    some_error = True
                    issues.append((3, "The Group specified for the parameter, '" + value + "', is not valid, in row " + str(r) + ". It must be a simple identifier."))
                param[k] = value
            elif k == "description":  # Optional
                param[k] = value

        # Check parameter completeness before adding it to the list of parameters
        if "name" not in param:
            issues.append((3, "The parameter must have a Name, row "+str(r)))
            continue
        if "type" not in param:
            issues.append((3, "The parameter must have a Type, row " + str(r)))
            continue

        content.append(param)

    return issues, None, content





from enum import Enum

from typing import List


class BasicEnum(Enum):
    @classmethod
    def get_names(cls) -> List[str]:
        return list(cls.__members__.keys())


class Scope(BasicEnum):
    Total = 1
    Internal = 2
    External = 3


class SubsystemType(BasicEnum):
    Local = (1, Scope.Internal)
    Environment = (2, Scope.Internal)
    External = (3, Scope.External)
    ExternalEnvironment = (4, Scope.External)

    def is_same_scope(self, other: 'SubsystemType') -> bool:
        return self.value[1] == other.value[1]

    def is_internal_scope(self) -> bool:
        return self.value[1] == Scope.Internal

    def is_external_scope(self) -> bool:
        return self.value[1] == Scope.External

    @staticmethod
    def from_str(s: str) -> 'SubsystemType':
        if s.lower() == 'local':
            return SubsystemType.Local
        elif s.lower() == 'environment':
            return SubsystemType.Environment
        elif s.lower() == 'external':
            return SubsystemType.External
        elif s.lower() == 'externalenvironment':
            return SubsystemType.ExternalEnvironment
        else:
            raise NotImplementedError




import json
from typing import Optional, Dict, Any

from nexinfosys.command_generators import Issue, IssueLocation, IType
from nexinfosys.common.helper import strcmp, first, PartialRetrievalDictionary
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import Observer, FactorTypesRelationUnidirectionalLinearTransformObservation, \
    FactorType, Processor, Indicator, IndicatorCategories, Hierarchy
from nexinfosys.command_executors import BasicCommand, CommandExecutionError, subrow_issue_message
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.models.musiasem_concepts_helper import find_or_create_observer, find_processor_by_name


class LCIAMethodsCommand(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))

    def _process_row(self, fields: Dict[str, Any], subrow=None) -> None:
        """
        :param fields:
        :param subrow:
        :return:
        """
        # InterfaceType must exist
        interface_type = self._glb_idx.get(FactorType.partial_key(fields["interface"]))
        if len(interface_type) == 1:
            pass
        elif len(interface_type) == 0:
            ft_name = fields["interface"]
            ft_h_name = "LCIA"
            hie = self._glb_idx.get(Hierarchy.partial_key(name=ft_h_name))
            if not hie:
                hie = Hierarchy(name=ft_h_name, type_name="interfacetype")
                self._glb_idx.put(hie.key(), hie)
            else:
                hie = hie[0]

            # Create it and warn
            ft = FactorType(ft_name,
                            parent=None, hierarchy=hie,
                            roegen_type=None,
                            tags=None,  # No tags
                            attributes=dict(unit=fields["interface_unit"], description=ft_name, level=None),
                            expression=None,
                            sphere=None,
                            opposite_processor_type=None
                            )
            # Simple name
            self._glb_idx.put(FactorType.partial_key(ft_name, ft.ident), ft)
            self._add_issue(IType.WARNING, f"InterfaceType with name '{ft_name}' not found, created" + subrow_issue_message(subrow))
        else:
            self._add_issue(IType.ERROR,
                            f"InterfaceType with name '{fields['interface']}' found {len(interface_type)} times" + subrow_issue_message(subrow))
            return

        auto_create = False

        # (LCIA) Indicator must exist
        indicator = self._glb_idx.get(Indicator.partial_key(fields["lcia_indicator"]))
        if len(indicator) == 1:
            pass
        elif len(indicator) == 0:
            if auto_create:
                # Create it and warn
                indicator = Indicator(fields["lcia_indicator"],
                                      f'LCIAMethod("{fields["lcia_indicator"]}")',
                                      None,
                                      "",
                                      [],
                                      IndicatorCategories.factors_expression,
                                      "",
                                      "LCIA",
                                      fields["interface_unit"],
                                      fields["interface_unit"],
                                      fields["lcia_method"])
                self._glb_idx.put(indicator.key(), indicator)
                self._add_issue(IType.WARNING, f"Indicator with name '{fields['lcia_indicator']}' not found, created" + subrow_issue_message(subrow))
            else:
                self._add_issue(IType.WARNING, f"Indicator with name '{fields['lcia_indicator']}' not found, skipped" + subrow_issue_message(subrow))
        else:
            self._add_issue(IType.ERROR,
                            f"Indicator with name '{fields['lcia_indicator']}' found {len(indicator)} times" + subrow_issue_message(subrow))
            return

        # Store LCIA Methods as a new variable.
        # TODO Use it to prepare a pd.DataFrame previous to calculating Indicators (after solving). Use "to_pickable"
        lcia_methods = self._state.get("_lcia_methods")
        if not lcia_methods:
            lcia_methods = PartialRetrievalDictionary()
            self._state.set("_lcia_methods", lcia_methods)
        horizon = fields.get("lcia_horizon", "")
        if horizon is None or horizon in ("_", "-"):
            horizon = ""
        compartment = fields.get("compartment", "")
        if compartment is None:
            compartment = ""
        subcompartment = fields.get("subcompartment", "")
        if subcompartment is None:
            subcompartment = ""
        _ = dict(m=fields["lcia_method"],
                 d=fields["lcia_indicator"],
                 h=horizon,
                 i=fields["interface"],
                 c=compartment,
                 s=subcompartment)
        from random import random
        # NOTE: a random() is generated just to grant that the value is unique
        lcia_methods.put(_, (fields["interface_unit"], fields["lcia_coefficient"], fields.get("lcia_interface_unit"), random()))




from typing import Dict, Any

from nexinfosys.command_executors import BasicCommand, subrow_issue_message
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.command_generators import IType
from nexinfosys.common.helper import create_dictionary
from nexinfosys.models.musiasem_concepts import Benchmark


class ScalarIndicatorBenchmarksCommand(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))

    def _process_row(self, fields: Dict[str, Any], subrow=None) -> None:
        """
        Create and register Benchmark object

        :param fields:
        """
        name = fields["benchmark"]
        benchmark_group = fields["benchmark_group"]
        stakeholders = fields["stakeholders"]
        b = self._glb_idx.get(Benchmark.partial_key(name=name))
        if len(b) == 1:
            b = b[0]
        elif len(b) == 0:
            b = Benchmark(name, benchmark_group, stakeholders.split(",") if stakeholders else [])
            self._glb_idx.put(b.key(), b)
        else:
            self._add_issue(IType.ERROR,
                            f"There are {len(b)} instances of the Benchmark '{name}'"+subrow_issue_message(subrow))
            return

        # Add range, if not repeated
        category = fields["category"]
        if category not in b.ranges:
            b.ranges[category] = create_dictionary(
                data=dict(range=fields["range"],
                          unit=fields["unit"],
                          category=category,
                          label=fields["label"],
                          description=fields["description"])
            )
        else:
            self._add_issue(IType.WARNING,
                            f"Range with category '{category}' repeated"+subrow_issue_message(subrow))




from nexinfosys.model_services import IExecutableCommand


class NestedCommandsCommand(IExecutableCommand):
    """
    This command is just a "placeholder" command, with no effect
    """
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        return None, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        return None  # NULLIFY CONTENT because commands have been serialized previously (FLAT APPROACH) self._content

    def json_deserialize(self, json_input):
        self._content = None
        return []



from io import StringIO
from typing import List

import pandas as pd
from openpyxl.worksheet.worksheet import Worksheet

from nexinfosys import IssuesLabelContentTripleType, AreaTupleType
from nexinfosys.command_generators import Issue, IssueLocation, IType
from nexinfosys.common.helper import strcmp, create_dictionary


def parse_dataset_data_command(sh: Worksheet, area: AreaTupleType, name: str, state) -> IssuesLabelContentTripleType:
    """
    Check that the syntax of the input spreadsheet is correct
    Return the analysis in JSON compatible format, for execution

    :param sh:   Input worksheet
    :param area: Area of the input worksheet to be analysed
    :return:     The command in a dict-list object (JSON ready)
    """

    issues: List[Issue] = []

    # Analyze column names
    col_map = create_dictionary()
    for c in range(area[2], area[3]):
        col_name = sh.cell(row=area[0], column=c).value.strip()
        # Avoid repetitions
        if col_name in col_map:
            issues.append(Issue(itype=IType.ERROR,
                                description="The column name '"+col_name+"' is repeated",
                                location=IssueLocation(sheet_name=name, row=1, column=c)))

        if strcmp(col_name, "DatasetName") or strcmp(col_name, "Dataset"):
            col_map["dataset"] = c
        elif col_name:
            # Concept name
            col_map[col_name] = c

    if any([i.itype == IType.ERROR for i in issues]):
        return issues, None, None

    # Read all the content into a list of lists
    lines = []
    for r in range(area[0] + 1, area[1]):
        line = []
        for col_name, c in col_map.items():
            v = sh.cell(row=r, column=c).value
            if isinstance(v, str):
                v = v.strip()
            line.append(v)
        lines.append(line)

    # pd.DataFrame
    df = pd.DataFrame(columns=[col_name for col_name in col_map], data=lines)

    content = []  # The output JSON

    if "dataset" in df:
        # Find the different datasets
        datasets = df["dataset"].unique()
        datasets = set([d.lower() for d in datasets])

        for dataset in datasets:
            # Obtain filtered
            df2 = df.loc[df['dataset'].str.lower() == dataset]
            # Convert to JSON and store in content
            del df2["dataset"]
            s = StringIO()
            df2.to_json(s, orient="split")
            content.append(dict(name=dataset, values=s.getvalue()))
    else:
        s = StringIO()
        df.to_json(s, orient="split")
        content.append(dict(name="", values=s.getvalue()))

    return issues, None, dict(items=content, command_name=name)

